from pathlib import Path
from typing import List, Optional, Dict, Union, Tuple, Literal, Sequence, Any
import fsspec
import numpy as np
from pandas.core.arrays import string_
from xarray import DataArray
from dataclasses import asdict, dataclass
import json
from fst.io.mrc import mrc_to_dask
from fst.io import read
import dask.array as da
import dacite
from xarray_multiscale.metadata.util import SpatialTransform

CONTAINER_TYPES ={'mrc', 'n5', 'precomputed'}
DTYPE_FORMATS = {"uint16": "n5", "uint8": "precomputed", "uint64": "n5"}
CONTENT_TYPES = {"em", "lm", "prediction", "segmentation", "analysis"}
ContainerTypes = Literal['n5', 'precomputed', 'mrc']


@dataclass
class VolumeStorageSpec:
    kvStore: str
    containerType: ContainerTypes
    containerPath: str
    dataPath: str

    def toURI(self):
        return f'{self.kvStore}://{Path(self.containerPath).with_suffix("." + self.containerType).joinpath(self.dataPath)}'

    def __post_init__(self):
        if self.containerType not in CONTAINER_TYPES:
            raise ValueError(
                f"containerType must be one of {CONTAINER_TYPES}"
            )

@dataclass
class ContrastLimits:
    min: float
    max: float

    def __post_init__(self):
        if not self.min <= self.max:
            raise ValueError('min must be less than or equal to max.')


@dataclass
class DisplaySettings:
    contrastLimits: ContrastLimits
    color: str = 'white'
    invertColormap: bool = False

    @classmethod
    def fromDict(cls, d: Dict[str, Any]):
        return dacite.from_dict(cls, d)


@dataclass 
class DatasetView:
    datasetName: str
    name: str
    description: str
    position: Optional[Sequence[float]]
    scale: Optional[float]
    volumeKeys: Sequence[str]
 
    @classmethod
    def fromDict(cls, d: Dict[str, Any]):
        return dacite.from_dict(cls, d)


@dataclass
class MultiscaleSpec:
    reduction: str
    depth: int
    factors: Union[int, Sequence[int]]


@dataclass
class MeshSource:
    path: str
    name: str
    datasetName: str
    format: str


@dataclass
class VolumeSource:
    path: str
    name: str
    datasetName: str
    dataType: str
    dimensions: Sequence[float]
    transform: SpatialTransform
    contentType: str
    containerType: Optional[ContainerTypes]
    displaySettings: DisplaySettings
    description: str = ''
    version: str="0"
    tags: Optional[Sequence[str]] = None

    def __post_init__(self):
        assert self.contentType in CONTENT_TYPES
        assert len(self.version) > 0

    def toDataArray(self):
        if Path(self.path).suffix == ".mrc":
            array = mrc_to_dask(self.path, chunks=(1, -1, -1))
        else:
            r = read(self.path)
            array = da.from_array(r, chunks=r.chunks)
        coords = [
            DataArray(
                self.transform.translate[idx] + np.arange(array.shape[idx]) * self.transform.scale[idx],
                dims=ax,
                attrs= {'units': self.transform.units[idx]}
                )
            for idx, ax in enumerate(self.transform.axes)
        ]
        return DataArray(array, coords=coords, name=self.name)


    @classmethod
    def fromDict(cls, d: Dict[str, Any]):
        return dacite.from_dict(cls, d)



@dataclass
class DatasetIndex:
    name: str
    volumes: Sequence[VolumeSource]
    meshes: Sequence[MeshSource]
    views: Sequence[DatasetView]
    
    @classmethod
    def from_json(cls, fname: Union[str, Path], open_kwargs: dict = {}):
        with fsspec.open(str(fname), mode='rt', **open_kwargs) as fh:
            jblob = json.loads(fh.read())
        return cls(**jblob)
    
    def to_json(self, fname: Union[str, Path], open_kwargs: dict = {}) -> int:
        jblob = json.dumps(asdict(self))
        with fsspec.open(str(fname), mode='wt', **open_kwargs) as fh:
            result = fh.write(jblob)
        return result


@dataclass
class VolumeIngest:
    source: VolumeSource
    multiscaleSpec: MultiscaleSpec
    storageSpec: VolumeStorageSpec
    mutation: Optional[str] = None


@dataclass
class COSEMArrayAttrs:
    name: str
    transform: SpatialTransform

    @classmethod
    def fromDataArray(cls, data: DataArray) -> "COSEMArrayAttrs":
        name = data.name
        if name is not None:
            return cls(str(name), SpatialTransform.fromDataArray((data)))
        else: 
            raise ValueError('DataArray argument must have a valid name')


@dataclass
class OMEScaleAttrs:
    path: str 
    transform: SpatialTransform


@dataclass
class OMEMultiscaleAttrs:
    datasets: Sequence[OMEScaleAttrs]


@dataclass
class COSEMGroupAttrs:
    name: str
    multiscales: Sequence[OMEMultiscaleAttrs]


@dataclass
class N5PixelResolution:
    dimensions: Sequence[float]
    unit: str


@dataclass
class NeuroglancerGroupAttrs:
    # see https://github.com/google/neuroglancer/issues/176#issuecomment-553027775
    axes: Sequence[str]
    units: Sequence[str]
    scales: Sequence[Sequence[int]]
    pixelResolution: N5PixelResolution


@dataclass
class MultiscaleGroupAttrs:
    name: str
    multiscales: Sequence[OMEMultiscaleAttrs]
    axes: Sequence[str]
    units: Sequence[str]
    scales: Sequence[Sequence[int]]
    pixelResolution: N5PixelResolution


def makeN5ArrayAttrs(dimensions: Sequence[float], unit: str) -> Dict[str, N5PixelResolution]:
    return {'pixelResolution': N5PixelResolution(dimensions, unit)}


def makeMultiscaleGroupAttrs(name: str,
                            arrays: Sequence[DataArray], 
                            array_paths: Sequence[str], 
                            axis_order: str="F") -> MultiscaleGroupAttrs:
    
    assert len(arrays) == len(array_paths)
    cosemArrayAttrs = tuple(COSEMArrayAttrs.fromDataArray(a) for a in arrays)
    
    axis_indexer = slice(None)
    # neuroglancer wants the axes reported in fortran order
    if axis_order == "F":
        axis_indexer = slice(-1, None, -1)
        
    axes: Tuple[str] = arrays[0].dims[axis_indexer]
    scales = tuple(tuple(s.scale_factors)[axis_indexer] for s in arrays)
    coords_reordered = tuple(arrays[0].coords[k] for k in axes)
    units = tuple(d.units for d in coords_reordered)

    # we need this for neuroglancer
    pixelResolution = N5PixelResolution(dimensions=cosemArrayAttrs[0].transform.scale[axis_indexer], unit=units[0])
    multiscales = OMEMultiscaleAttrs(datasets=[OMEScaleAttrs(path=ap, transform=attr.transform) for ap, attr in zip(array_paths, cosemArrayAttrs)])

    result = MultiscaleGroupAttrs(name=name,
                                  multiscales=[multiscales], 
                                  axes=axes,
                                  units=units,
                                  scales=scales,
                                  pixelResolution=pixelResolution)
    return result


@dataclass
class CompositeArrayAttrs:
    name: str
    transform: SpatialTransform
    pixelResolution: N5PixelResolution

    @classmethod
    def fromDataArray(cls, data: DataArray):
        cosemAttrs = COSEMArrayAttrs.fromDataArray(data)
        pixelResolution = N5PixelResolution(cosemAttrs.transform.scale[::-1], unit=cosemAttrs.transform.units[0])
        return cls(cosemAttrs.name, cosemAttrs.transform, pixelResolution)


from pathlib import Path
import argparse
import subprocess
import os

ext = ".dat"


def main():

    parser = argparse.ArgumentParser(
        description="Save the extended metadata from the footer of a .dat file as json and xml "
        "files in the same directory as the .dat file. This script exists to "
        "work around a the fact that certain important metadata in .dat files is "
        " in an undocumented region in the end of the file. This script uses a "
        "LabVIEW executible written by Dan Milkie to serialize the metadata."
    )

    parser.add_argument(
        "-s",
        "--data_root",
        help="Path to a directory containing directories containing .dat files",
        required=True,
    )

    parser.add_argument(
        "-e",
        "--exe_path",
        help="Path to the executible which extracts the metadata from the .dat files",
        required=True,
    )

    args = parser.parse_args()
    if not os.path.exists(args.data_root):
        raise FileNotFoundError(
            f"The data root path {args.data_root} could not be found."
        )
    elif not os.path.exists(args.exe_path):
        raise FileNotFoundError(f"The exe path {args.exe_path} could not be found.")
    toProcess = []
    rawDirs = Path(args.data_root).glob(f"*/raw/")
    toProcess.extend([str(list(sorted(x.glob(f"*{ext}")))[0]) for x in rawDirs])
    cmds = [[args.exe_path, t] for t in toProcess]
    [subprocess.run(c) for c in cmds]


if __name__ == "__main__":
    main()



from dataclasses import asdict
from .fibsem import read_fibsem
from pathlib import Path
from typing import (
    Union,
    Iterable,
    List,
    Optional,
    Callable,
    Dict,
    Tuple,
    Sequence,
    Any,
)
from dask import delayed
import dask.array as da
import os
from shutil import rmtree
from itertools import groupby, chain
from collections import defaultdict
from dask.diagnostics import ProgressBar
import zarr
import h5py
import dask
from mrcfile.mrcmemmap import MrcMemmap
from xarray import DataArray
from xarray.core.coordinates import DataArrayCoordinates
import numpy as np
from dask import bag
from zarr.core import Array as ZarrArray
from zarr.hierarchy import Group as ZarrGroup
from .mrc import mrc_shape_dtype_inference, access_mrc, mrc_to_dask
from numcodecs import GZip

# encode the fact that the first axis in zarr is the z axis
_zarr_axes = {"z": 0, "y": 1, "x": 2}
# encode the fact that the first axis in n5 is the x axis
_n5_axes = {"z": 2, "y": 1, "x": 0}
_formats = (".dat", ".mrc")
_container_extensions = (".zarr", ".n5", ".h5", ".precomputed")
_suffixes = (*_formats, *_container_extensions)

Pathlike = Union[str, Path]
Arraylike = Union[zarr.core.Array, da.Array, DataArray, np.array]
ArraySources = Union[
    List[dask.delayed], zarr.core.Array, zarr.hierarchy.Group, h5py.Dataset, h5py.Group, np.array
]

defaultUnit = "nm"


def broadcast_kwargs(**kwargs) -> Dict:
    """
    For each keyword: arg in kwargs, assert that there are only 2 types of args: sequences with length = 1 
    or sequences with some length = k. Every arg with length 1 will be repeated k times, such that the return value 
    is a dict of kwargs with minimum length = k.
    """
    grouped: Dict[str, List] = defaultdict(list)
    sorter = lambda v: len(v[1])
    s = sorted(kwargs.items(), key=sorter)
    for l, v in groupby(s, key=sorter):
        grouped[l].extend(v)

    assert len(grouped.keys()) <= 2
    if len(grouped.keys()) == 2:
        assert min(grouped.keys()) == 1
        output_length = max(grouped.keys())
        singletons, nonsingletons = tuple(grouped.values())
        singletons = ((k, v * output_length) for k, v in singletons)
        result = {**dict(singletons), **dict(nonsingletons)}
    else:
        result = kwargs

    return result


def split_path_at_suffix(
    upper_path: Pathlike, lower_path: Pathlike = "", suffixes: tuple = _suffixes
) -> List[Path]:
    """
    Recursively climb a path, checking at each level of the path whether the tail of the path represents a directory
    with a container extension. Returns the path broken at the level where a container is found.  
    """
    upper, lower = Path(upper_path), Path(lower_path)

    if upper.suffix in suffixes:
        result = [upper, lower]
    else:
        if len(upper.parts) >= 2:
            result = split_path_at_suffix(
                Path(*upper.parts[:-1]), Path(upper.parts[-1], lower), suffixes
            )
        else:
            raise ValueError(
                f"Could not find any suffixes matching {suffixes} in {upper / lower}"
            )

    return result


def access_fibsem(path: Union[Pathlike, Iterable[str], Iterable[Path]], mode: str):
    if mode != "r":
        raise ValueError(
            f".dat files can only be accessed in read-only mode, not {mode}."
        )
    return read_fibsem(path)


def access_n5(
    dir_path: Pathlike, container_path: Pathlike, **kwargs
) -> Union[zarr.core.Array, zarr.hierarchy.Group]:
    return zarr.open(zarr.N5Store(dir_path), path=container_path, **kwargs)


def access_zarr(
    dir_path: Pathlike, container_path: Pathlike, **kwargs
) -> Union[zarr.core.Array, zarr.hierarchy.Group]:
    return zarr.open(str(dir_path), path=str(container_path), **kwargs)


def access_h5(
    dir_path: Pathlike, container_path: Pathlike, mode: str, **kwargs
) -> Union[h5py.Dataset, h5py.Group]:
    result = h5py.File(dir_path, mode=mode, **kwargs)
    if container_path != "":
        result = result[str(container_path)]
    return result


def access_precomputed(outer_path: Pathlike, 
                       inner_path: Pathlike, 
                       mode: str):

    from .tensorstore import TensorStoreSpec, KVStore, parse_info
    import tensorstore as ts

    parent_dir = str(Path(outer_path).parent)
    container_dir = Path(outer_path).parts[-1]    
    kvstore = KVStore(driver='file', path=parent_dir)
    
    if mode == 'r':
        # figure out what arrays already exist within the container
        info = parse_info((Path(outer_path) / 'info').read_text())        
        scale_keys = [s.key for s in info.scales]
        if len(str(inner_path)) > 0:          
            # find the scale index corresponding to the inner path
            scale_index = scale_keys.index(str(inner_path))
        else:
            scale_index = 0

        spec = TensorStoreSpec(driver='neuroglancer_precomputed', 
                                kvstore=kvstore,
                                path=str(container_dir), 
                                scale_index=scale_index)

        result = ts.open(spec=spec.asdict(), read=True).result()
    
    elif mode == 'w':
        result = None
    
    return result

accessors: Dict[str, Callable] = {}
accessors[".dat"] = access_fibsem
accessors[".n5"] = access_n5
accessors[".zarr"] = access_zarr
accessors[".h5"] = access_h5
accessors[".mrc"] = access_mrc
accessors['.precomputed'] = access_precomputed


def access(
    path: Union[Pathlike, Iterable[str], Iterable[Path]],
    mode: str,
    lazy: bool = False,
    **kwargs,
) -> ArraySources:
    """

    Access data from a variety of array storage formats.

    Parameters
    ----------
    path: A path or collection of paths to image files. If `path` is a string, then the appropriate reader will be
          selected based on the extension of the path, and the file will be read. If `path` is a collection of strings,
          it is assumed that each string is a path to an image and each will be read sequentially.

    lazy: A boolean, defaults to False. If True, this function returns the native file reader wrapped by
    dask.delayed. This is advantageous for distributed computing.

    mode: The access mode for the file. e.g. 'r' for read-only access.

    Returns an array-like object, a collection of array-like objects, a chunked store, or
    a dask.delayed object.
    -------

    """
    if isinstance(path, (str, Path)):
        path_inner: Pathlike
        path_outer, path_inner = split_path_at_suffix(path)

        # str(Path('')) => '.', which we don't want for an empty trailing path
        if str(path_inner) == ".":
            path_inner = ""

        fmt = path_outer.suffix
        is_container = fmt in _container_extensions

        try:
            accessor = accessors[fmt]
        except KeyError:
            raise ValueError(
                f"Cannot access images with extension {fmt}. Try one of {list(accessors.keys())}"
            )

        if lazy:
            accessor = delayed(accessor)
        if is_container:
            return accessor(path_outer, path_inner, mode=mode, **kwargs)
        else:
            return accessor(path_outer, mode=mode, **kwargs)

    elif isinstance(path, Iterable):
        return [access(p, mode, lazy, **kwargs) for p in path]
    else:
        raise ValueError("`path` must be a string or iterable of strings")


def read(path: Union[Pathlike, Iterable[str], Iterable[Path]], lazy=False, **kwargs):
    """

    Access data on disk with read-only permissions

    Parameters
    ----------
    path: A path or collection of paths to image files. If `path` is a string, then the appropriate image reader will be
          selected based on the extension of the path, and the file will be read. If `path` is a collection of strings,
          it is assumed that each string is a path to an image and each will be read sequentially.

    lazy: A boolean, defaults to False. If True, this function returns the native file reader wrapped by
    dask.delayed. This is advantageous for distributed computing.

    Returns an array-like object, a collection of array-like objects, a chunked store, or
    a dask.delayed object.
    -------

    """
    return access(path, mode="r", lazy=lazy, **kwargs)


def infer_coordinates_3d(arr: Arraylike, default_unit="nm") -> List[DataArray]:
    """
    Infer the coordinates and units from a 3D volume.

    """

    units: Dict[str, str]
    coords: List[DataArray]
    scaleDict: Dict[str, float]

    assert arr.ndim == 3

    # DataArray: get coords attribute directly from the data
    if hasattr(arr, "coords"):
        coords = arr.coords

    # zarr array or hdf5 array: get coords from attrs
    elif hasattr(arr, "attrs"):
        if arr.attrs.get("pixelResolution") or arr.attrs.get("resolution"):
            if pixelResolution := arr.attrs.get("pixelResolution"):
                scale: List[float] = pixelResolution["dimensions"]
                unit: str = pixelResolution["unit"]
            elif scale := arr.attrs.get("resolution"):
                unit = default_unit

            scaleDict = {k: scale[v] for k, v in _n5_axes.items()}
        else:
            scaleDict = {k: 1 for k, v in _n5_axes.items()}
            unit = default_unit

        coords = [
            DataArray(
                np.arange(arr.shape[v]) * scaleDict[k], dims=k, attrs={"units": unit}
            )
            for k, v in _zarr_axes.items()
        ]

    else:
        # check if this is a dask array constructed from a zarr array
        if isinstance(arr, da.Array) and isinstance(get_array_original(arr), zarr.core.Array):
            arr_source: zarr.core.Array = get_array_original(arr)
            coords = infer_coordinates_3d(arr_source)
        else:
            coords = [
                DataArray(np.arange(arr.shape[v]), dims=k, attrs={"units": defaultUnit})
                for k, v in _zarr_axes.items()
            ]

    return coords


def DataArrayFromFile(source_path: Pathlike):
    if Path(source_path).suffix == '.mrc':
        arr = mrc_to_dask(source_path, chunks=(1,-1,-1)) 
    else:
        arr = read(source_path)
        arr = da.from_array(arr, chunks=arr.chunks)
        if not hasattr(arr, "chunks"):
            raise ValueError(
                f'{arr} does not have a "chunks" attribute. Is it really a distributed array-like?'
            )
    coords = infer_coordinates_3d(arr)
    data = DataArrayFactory(
        arr,
        attrs={"source": str(source_path)},
        coords=coords,
    )
    return data


def DataArrayFactory(arr: Arraylike, **kwargs):
    """
    Create an xarray.DataArray from an array-like input (e.g., zarr array, dask array). This is a very light 
    wrapper around the xarray.DataArray constructor that checks for cosem/n5 metadata attributes and uses those to
    generate DataArray.coords and DataArray.dims properties; additionally, metadata about units will be inferred and 
    inserted into the `attrs` kwarg if it is supplied.

    Parameters
    ----------

    arr: Array-like object (dask array or zarr array)

    """
    attrs: Optional[Dict]
    extra_attrs = {}

    # if we pass in a zarr array, daskify it first
    # maybe later add hdf5 support here
    if isinstance(arr, zarr.core.Array):
        source = str(Path(arr.store.path) / arr.path)
        # save the full path to the array as an attribute
        extra_attrs["source"] = source
        arr = da.from_array(arr, chunks=arr.chunks)

    coords = infer_coordinates_3d(arr)
    if "coords" not in kwargs:
        kwargs.update({"coords": coords})

    if attrs := kwargs.get("attrs"):
        out_attrs = attrs.copy()
        out_attrs.update(extra_attrs)
        kwargs["attrs"] = out_attrs
    else:
        kwargs["attrs"] = extra_attrs

    data = DataArray(arr, **kwargs)
    return data


def create_array(group, attrs=None, **kwargs) -> zarr.core.Array:
    name = kwargs["name"]
    overwrite = kwargs.get("overwrite", False)
    if name not in group:
        arr = group.zeros(**kwargs)
    else:
        arr = group[name]
        if not same_array_props(
            arr,
            shape=kwargs["shape"],
            dtype=kwargs["dtype"],
            compressor=kwargs["compressor"],
            chunks=kwargs["chunks"],
        ):
            arr = group.zeros(**kwargs)
        else:
            if overwrite == False:
                raise FileExistsError(
                    f"{group.path}/{name} already exists as an array. Call this function with overwrite=True to delete this array."
                )
    if attrs is not None:
        arr.attrs.put(attrs)
    return arr


def create_arrays(
    path: Union[str, Path],
    names: Sequence[str],
    shapes: Sequence[Sequence[int]],
    dtypes: Sequence[str],
    compressors: Sequence[Any],
    chunks: Sequence[Sequence[int]],
    group_attrs: Dict[str, Any] = {},
    attrs: Optional[Sequence[Optional[Dict[str, Any]]]] = None,
    overwrite: bool = True,
) -> Tuple[zarr.hierarchy.Group, Tuple[zarr.core.Array]]:
    """
    Use Zarr / N5 to create a collection of arrays within a group (the group will also be created, if needed). If overwrite==True,
    these arrays will be created as needed and filled with 0s. Otherwise, new arrays will be created, existing arrays with matching properties
    will be kept as-is, and existing arrays with mismatched properties will be removed and replaced with an array of 0s.      
    """

    group: zarr.hierarchy.Group = access(path, mode="a")
    group.attrs.put(group_attrs)

    if attrs is None:
        attrs = [None] * len(names)

    delayed_creator = delayed(create_array)

    arrs_delayed: List[Any] = []
    for ind in range(len(names)):
        arrs_delayed.append(
            delayed_creator(
                group=group, 
                attrs=attrs[ind], 
                name=names[ind], 
                shape=shapes[ind],
                dtype=dtypes[ind],
                compressor=compressors[ind],
                chunks=chunks[ind],
                overwrite=overwrite))

    with ProgressBar():
        arrs: Tuple[zarr.core.Array] = delayed(arrs_delayed).compute(scheduler="threads")

    return group, arrs


def access_multiscale(
    container_path: Pathlike,
    group_path: Pathlike,
    arrays: Sequence[DataArray],
    array_paths: Sequence[str], 
    array_chunks: Sequence[int],
    root_attrs: Optional[Dict[str, Any]]=None,
    group_attrs: Dict[str, Any] = {},
    compressor: Any=GZip(-1),
    attr_factory: Optional[Callable[[DataArray], Dict[str, Any]]] = None
) -> Tuple[zarr.hierarchy.group, zarr.Array]:

    root: zarr.hierarchy.Group = access(container_path, mode="a")
    if root_attrs is not None:
        root.attrs.put(root_attrs)
    
    if attr_factory is None:
        attr_factory = lambda v: {}

    shapes, dtypes, compressors, chunks, arr_attrs = [], [], [], [], []
    for p in arrays:
        shapes.append(p.shape)
        dtypes.append(p.dtype)
        compressors.append(compressor)
        chunks.append(array_chunks)
        arr_attrs.append(attr_factory(p))

    zgrp, zarrays = create_arrays(
        Path(container_path) / group_path,
        names=array_paths,
        shapes=shapes,
        dtypes=dtypes,
        compressors=compressors,
        chunks=chunks,
        group_attrs=group_attrs,
        attrs=arr_attrs,
    )

    return zgrp, zarrays


def rmtree_parallel(path: Pathlike) -> int:
    """
    Recursively remove the contents of a directory in parallel. All files are found using os.path.walk, then dask 
    is used to delete the files in parallel. Finally, the (empty) directories are removed.

    path: String, a path to the container folder, e.g. /home/user/tmp/

    return: 0

    """
    # find all the files using os.walk
    files = fwalk(path)
    if len(files) > 0:
        bg = bag.from_sequence(files)
        bg.map(lambda v: os.remove(v)).compute(scheduler="processes")
    rmtree(path)
    return 0


def same_compressor(arr: zarr.Array, compressor) -> bool:
    """

    Determine if the compressor associated with an array is the same as a different compressor.

    arr: A zarr array
    compressor: a Numcodecs compressor, e.g. GZip(-1)
    return: True or False, depending on whether the zarr array's compressor matches the parameters (name, level) of the
    compressor.
    """
    comp = arr.compressor.compressor_config
    return comp["id"] == compressor.codec_id and comp["level"] == compressor.level


def same_array_props(
    arr: zarr.Array, shape: Tuple[int], dtype: str, compressor: Any, chunks: Tuple[int]
) -> bool:
    """

    Determine if a zarr array has properties that match the input properties.

    arr: A zarr array
    shape: A tuple. This will be compared with arr.shape.
    dtype: A numpy dtype. This will be compared with arr.dtype.
    compressor: A numcodecs compressor, e.g. GZip(-1). This will be compared with the compressor of arr.
    chunks: A tuple. This will be compared with arr.chunks
    return: True if all the properties of arr match the kwargs, False otherwise.
    """
    return (
        (arr.shape == shape)
        & (arr.dtype == dtype)
        & same_compressor(arr, compressor)
        & (arr.chunks == chunks)
    )


def get_array_original(arr: da.Array) -> Any:
    """
    Return the zarr array that was used to create a dask array using `da.from_array(zarr_array)`
    """
    keys = tuple(arr.dask.keys())
    return arr.dask[keys[-1]]


def fwalk(source: Pathlike, endswith: Union[str, Tuple[str, ...]] = "") -> List[str]:
    """
    Use os.walk to recursively parse a directory tree, returning a list containing the full paths
    to all files with filenames ending with `endswith`.
    """
    results = []
    for p, d, f in os.walk(source):
        for file in f:
            if file.endswith(endswith):
                results.append(os.path.join(p, file))
    return results


def fwalk_parallel(elements: Sequence[Path]):
    """
    Given a sequence containing either files or directories, separate the input files, walk the directories,
    in parallel, and return the all files found + input files 
    """
    files, dirs = [], []
    for k,v in groupby(sorted(elements, key=Path.is_dir), key=Path.is_dir):
        if k: dirs.extend([delayed(fwalk)(val) for val in list(v)])
        else: files.extend(list(map(str, v)))
    dirs_computed = delayed(dirs).compute()
    files.extend(list(chain(*dirs_computed)))
    return  files


def infer_dtype(path: str) -> str:
    fd = read(path)
    if hasattr(fd, "dtype"):
        dtype = str(fd.dtype)
    elif hasattr(fd, "data"):
        _, dtype = mrc_shape_dtype_inference(fd)
        dtype=str(dtype)
    else:
        raise ValueError(f"Cannot infer dtype of data located at {path}")
    return dtype


from os import name

from typing import Sequence
from dataclasses import dataclass
import xarray
import numpy as np
from pathlib import Path
from typing import Optional, Dict, List, Tuple, Union, Mapping
from ..io.io import infer_dtype
from ..io import read, DataArrayFromFile
from dataclasses import asdict
from ..attrs import (
    VolumeSource,
    DisplaySettings,
    ContrastLimits,
    DatasetView,
    MeshSource,
)
from xarray_multiscale.metadata.util import SpatialTransform
from pymongo import MongoClient, ReplaceOne
from sheetscrape.scraper import GoogleSheetScraper
import pandas as pd
from fst.attrs import DatasetView
import zarr
import click

# for volume names, which underscore-separated trailing values are suffixes:
# e.g., 'mito_seg' has 'seg' as a suffix, while 'mito_er_contacts' has no suffix and is itself an entire token (for now)
volume_name_suffixes = ('seg', 'pred')

# constants for getting stuff from google docs
credfile = '/groups/scicompsoft/home/bennettd/certificates/cosem-db-1669c2970c7f.json'
view_sheet_name = 'FIB-SEM dataset regions of interest'
view_sheet_page = 0


def dir_glob(path: str) -> Tuple[str, ...]:
    return tuple(sorted(map(str, Path(path).glob("*"))))


def make_segmentation_display_settings(path: str, prediction_colors: Dict[str, str]):
    return DisplaySettings(
        contrastLimits=ContrastLimits(0, 1), color=prediction_colors[Path(path).name]
    )

# groundtruth is .5 x the grid spacing of raw data
ground_truth_scaling = 0.5
base_res = 4.0
yx_res = {"y": base_res, "x": base_res}
base_display = DisplaySettings(ContrastLimits(0, 1.0), "white", True)

axes = ("z", "y", "x")
units = ("nm",) * len(axes)
translate = (0,) * len(axes)


def scale(
    factors: Union[float, Sequence[float]], transform: SpatialTransform
) -> SpatialTransform:
    return SpatialTransform(
        transform.axes,
        transform.units,
        transform.translate,
        tuple(np.multiply(factors, transform.scale)),
    )

def infer_container_type(path: str) -> str:
    if Path(path).suffix == ".mrc":
        containerType = "mrc"
    elif any(map(lambda v: (Path(v).suffix == ".n5"), Path(path).parts)):
        containerType = "n5"
    elif any(map(lambda v: (Path(v).suffix == ".precomputed"), Path(path).parts)):
        containerType = "precomputed"
    else:
        raise ValueError(f"Could not infer container type from path {path}")

    return containerType

# todo: fix this generating the wrong info for analysis volumes
def get_classname_and_content_type(volume_name: str, suffixes: Sequence[str] = volume_name_suffixes) -> Tuple[str, str]:
    # remove all whitespace before splitting on _
    separated = ''.join(volume_name.split()).split("_")
    if separated[-1] in suffixes:
        class_name = separated[0]
    else:
        class_name = volume_name 
    content_type = ""
    if separated[-1] == "pred":
        content_type = "prediction"
    elif separated[-1] == "seg":
        content_type = "segmentation"
    else:
        content_type = "analysis"
    return class_name, content_type


base_transforms = {
    "jrc_hela-4": SpatialTransform(
        axes, units, translate, (base_res * 1.07, base_res, base_res)
    ),
    "jrc_mus-pancreas-1": SpatialTransform(
        axes, units, translate, (base_res * 0.85, base_res, base_res)
    ),
    "jrc_hela-2": SpatialTransform(
        axes, units, translate, (base_res * 1.31, base_res, base_res)
    ),
    "jrc_hela-3": SpatialTransform(
        axes, units, translate, (base_res * 0.81, base_res, base_res)
    ),
    "jrc_jurkat-1": SpatialTransform(
        axes, units, translate, (base_res * 0.86, base_res, base_res)
    ),
    "jrc_macrophage-2": SpatialTransform(
        axes, units, translate, (base_res * 0.84, base_res, base_res)
    ),
    "jrc_sum159-1": SpatialTransform(
        axes, units, translate, (base_res * 1.14, base_res, base_res)
    ),
    "jrc_ctl-id8-1": SpatialTransform(
        axes, units, translate, (base_res * 0.87, base_res, base_res)
    ),
    "jrc_fly-fsb-1": SpatialTransform(
        axes, units, translate, (base_res * 1.0, base_res, base_res)
    ),
    "jrc_fly-acc-calyx-1": SpatialTransform(
        axes, units, translate, (base_res * 0.93, base_res, base_res)
    ),
    "jrc_hela-1": SpatialTransform(axes, units, translate, (base_res * 2,) * 3),
    "jrc_choroid-plexus-2": SpatialTransform(
        axes, units, translate, (base_res * 2,) * 3
    ),
    "jrc_cos7-11": SpatialTransform(axes, units, translate, (base_res * 2,) * 3),
}

class_info = {
    "cent": ("Centrosome", "white"),
    "cent-dapp": ("Centrosome Distal Appendage", "white"),
    "cent-sdapp": ("Centrosome Subdistal Appendage", "white"),
    "chrom": ("Chromatin", "white"),
    "er": ("Endoplasmic Reticulum", "blue"),
    "er_palm" : ("Light Microscopy (PALM) of ER", "white"),
    "er_sim" : ("Light Microscopy (SIM) of the ER", "white"), 
    "er-mem": ("Endoplasmic Reticulum membrane", "blue"),
    "eres": ("Endoplasmic Reticulum Exit Site", "white"),
    "eres-mem": ("Endoplasmic Reticulum Exit Site membrane", "white"),
    "endo": ("Endosomal Network", "magenta"),
    "endo-mem": ("Endosome membrane", "magenta"),
    "echrom": ("Euchromatin", "white"),
    "ecs": ("Extracellular Space", "white"),
    "golgi": ("Golgi", "cyan"),
    "golgi-mem": ("Golgi membrane", "cyan"),
    "hchrom": ("Heterochromatin", "white"),
    "ld": ("Lipid Droplet", "white"),
    "ld-mem": ("Lipid Droplet membrane", "white"),
    "lyso": ("Lysosome", "white"),
    "lyso-mem": ("Lysosome membrane", "white"),
    "mt": ("Microtubule", "orange"),
    "mt-in": ("Microtubule inner", "orange"),
    "mt-out": ("Microtubule outer", "orange"),
    "mito": ("Mitochondria", "green"),
    "mito_palm" : ("Light Microscopy (PALM) of Mitochondria", "white"),
    "mito_sim" : ("Light Microscopy (SIM) of Mitochondria", "white"), 
    "mito-mem": ("Mitochondria membrane", "green"),
    "mito-ribo": ("Mitochondria Ribosome", "white"),
    "ne": ("Nuclear Envelope", "white"),
    "ne-mem": ("Nuclear Envelope membrane", "white"),
    "np": ("Nuclear Pore", "white"),
    "np-in": ("Nuclear Pore inner", "white"),
    "np-out": ("Nuclear Pore outer", "white"),
    "nucleolus": ("Nucleoulus", "white"),
    "nechrom": ("Nucleoulus associated Euchromatin", "white"),
    "nhchrom": ("Nucleoulus associated Heterochromatin", "white"),
    "nucleus": ("Nucleus", "red"),
    "pm": ("Plasma Membrane", "orange"),
    "ribo": ("Ribosome", "yellow"),
    "vesicle": ("Vesicle", "red"),
    "vesicle-mem": ("Vesicle membrane", "red"),
    "er_ribo_contacts": ("ER - Ribosome Contact Sites", "white"),
    "er_golgi_contacts": ("ER - Golgi Contact Sites", "white"),
    "er_mito_contacts": ("ER - Mito Contact Sites", "white"),
    "endo_er_contacts": ("Endosome - ER Contact Sites", "white"),
    "er_nucleus_contacts": ("ER - Nucleus Contact Sites", "white"),
    "er_pm_contacts": ("ER - Plasma Membrane Contat Sites", "white"),
    "er_vesicle_contacts": ("ER - Vesicle Contact Sites", "white"),
    "golgi_vesicle_contacts": ("Golgi - Vesicle Contact Sites", "white"),
    "endo_golgi_contacts": ("Endosome - Golgi Contact Sites", "white"),
    "mito_pm_contacts": ("Mito - Plasma Membrane Contact Sites", "white"),
    "er_mt_contacts": ("ER - Microtubule Contact Sites", "white"),
    "endo_mt_contacts": ("Endosome - Microtubule Contact Sites", "white"),
    "golgi_mt_contacts": ("Golgi - Microtubule Contact Sites", "white"),
    "mito_mt_contacts": ("Mitochondria - Microtubule Contact Sites", "white"),
    "mt_nucleus_contacts": ("Microtubule - Nucleus Contact Sites", "white"),
    "mt_vesicle_contacts": ("Microtubule - Vesicle Contact Sites", "white"),
    "mt_pm_contacts": ("Microtubule - Plasma Membrane Contact Sites", "white"),
    "mito_skeleton": ("Mitochrondria Skeletons", "white"),
    "mito_skeleton-lsp": ("Mitochrondria Skeletons: Longest Shortest Path", "white"),
    "er_medial-surface": ("ER Medial Surface", "white"),
    "er_curvature": ("Reconstructed ER from Medial Surface with Curvature", "white"),
    'ribo_classified' : ('Ribosomes classified by contact surface', 'white'),
    "fibsem-uint8": ("FIB-SEM Data (compressed)", "white"),
    "fibsem-uint16": ("FIB-SEM Data (uncompressed)", "white"),
    "gt": ("Ground truth", "white"),
}



def get_views_from_google_docs(credfile, spreadsheet_name, spreadsheet_page):
    sheet = GoogleSheetScraper(credfile, mode='r').client.open(spreadsheet_name)
    # convert the sheet to a dataframe
    columns, *values = sheet.worksheets()[spreadsheet_page].get_all_values()
    df = pd.DataFrame(values, columns=columns)
    return views_from_dataframe(df)

def views_from_dataframe(df: pd.DataFrame):
    views = []
    for idx,r in df.iterrows():
        datasetName: str = r['Dataset name']
        name: str = r['ROI Name (short)']
        description: str = r['ROI Description (long)']
        pos_str: str = r['XYZ Coordinates (nm)']
        pos: Optional[Sequence[int]]
        scale_str: str = r['Cross-Section Scale']
        scale: Optional[float]

        if len(pos_str) > 0:
            pos = tuple(map(int, ''.join(pos_str.split()).split(',')))
        else:
            pos = None
        
        if scale_str == '':
            scale = None
        else:
            scale = float(scale_str)

        layers: List[str] = r['Layers'].split(', ')
        for l in layers:
            class_name, _ = get_classname_and_content_type(l)
            if class_name not in class_info:
                print(f'Warning: layer name {l} from the view called {name} was not found in the list of layers.')
        views.append(DatasetView(datasetName, name, description, pos, scale, layers))
    return views


@dataclass
class RawSources:
    datasetName: str
    uint8: Optional[Tuple[str, DisplaySettings]] = None
    uint16: Optional[Tuple[str, DisplaySettings]] = None
    pred: Optional[Tuple[str, ...]] = None
    groundTruth: Optional[str] = None
    meshes: Tuple[str, ...] = ()
    lm: Optional[Tuple[str, ...]] = None


hess_raw_dir = "/groups/hess/hesslab/HighResPaper_rawdata"
cosem_raw_dir = "/groups/cosem/cosem/data"
raw_sources = (
    RawSources(
        "jrc_hela-4",
        (
            f"{hess_raw_dir}/10. HeLa_mitotic_17-7_Cell4_4x4x4nm/HeLa_mitotic_17-7_17_Cell4 4x4x4nm_raw.mrc",
            DisplaySettings(ContrastLimits(0.57, 0.805)),
        ),
        (
            f"{hess_raw_dir}/10. HeLa_mitotic_17-7_Cell4_4x4x4nm/HeLa_mitotic_17-7_17_Cell4 4x4x4nm 16bit.mrc",
            DisplaySettings(ContrastLimits(0.275, 0.355), invertColormap=True),
        ),
    ),
    RawSources(
        "jrc_mus-pancreas-1",
        (
            f"{hess_raw_dir}/1. Pancreas Islet_G64-2-1-HighGlucose_4x4x4nm/G64-2-1_HighGlucose 4x4x4nm.mrc",
            DisplaySettings(ContrastLimits(0.7176, 0.8117)),
        ),
        (
            f"{hess_raw_dir}/1. Pancreas Islet_G64-2-1-HighGlucose_4x4x4nm/G64-2-1_HighGlucose 4x4x4nm 16bit.mrc",
            DisplaySettings(ContrastLimits(0.108, 0.199), invertColormap=True),
        ),
    ),
    RawSources(
        "jrc_hela-2",
        (
            f"{cosem_raw_dir}/HeLa_Cell2_4x4x4nm/Aubrey_17-7_17_Cell2 4x4x4nm.mrc",
            DisplaySettings(ContrastLimits(0.415, 0.716)),
        ),
        (
            f"{hess_raw_dir}/2. HeLa2_Aubrey_17-7_17_Cell2_4x4x4nm/Aubrey_17-7_17_Cell2 4x4x4nm 16bit.mrc",
            DisplaySettings(ContrastLimits(0.373, 0.52), invertColormap=True),
        ),
        pred=dir_glob(
            "/groups/cosem/cosem/ackermand/forDavis/jrc_hela-2/jrc_hela-2.n5"
        ),
        groundTruth="/nrs/cosem/bennettd/groundtruth/jrc_hela-2/jrc_hela-2.n5/groundtruth_0003/",
        meshes=dir_glob('/nrs/cosem/bennettd/s3/janelia-cosem/jrc_hela-2/neuroglancer/mesh')
    ),
    RawSources(
        "jrc_hela-3",
        (
            f"{hess_raw_dir}/3. HeLa3_Aubrey_17-7_17_Cell3_4x4x4nm/HeLa_Cell3_17-7_17_4x4x4nm.mrc",
            DisplaySettings(ContrastLimits(0.216, 0.944)),
        ),
        (
            f"{hess_raw_dir}/3. HeLa3_Aubrey_17-7_17_Cell3_4x4x4nm/HeLa_Cell3_17-7_17_4x4x4nm 16bit.mrc",
            DisplaySettings(ContrastLimits(0.313, 0.409), invertColormap=True),
        ),
        pred=dir_glob(
            "/groups/cosem/cosem/ackermand/forDavis/jrc_hela-3/jrc_hela-3.n5"
        ),
        groundTruth="/nrs/cosem/bennettd/groundtruth/jrc_hela-3/jrc_hela-3.n5/groundtruth_0003/",
    ),
    RawSources(
        "jrc_jurkat-1",
        (
            f"{hess_raw_dir}/7. Jurkat_Cryo_2017_FS96_Cell1_4x4x4nm/Jurkat_Cryo_2017_FS96_Area1 4x4x4nm.mrc",
            DisplaySettings(ContrastLimits(0.826, 0.924)),
        ),
        (
            f"{hess_raw_dir}/7. Jurkat_Cryo_2017_FS96_Cell1_4x4x4nm/Jurkat_Cryo_2017_FS96_Area1 4x4x4nm 16bit.mrc",
            DisplaySettings(ContrastLimits(0.297, 0.402), invertColormap=True),
        ),
        pred=dir_glob(
            "/groups/cosem/cosem/ackermand/forDavis/jrc_jurkat-1/jrc_jurkat-1.n5"
        ),
        groundTruth="/nrs/cosem/bennettd/groundtruth/jrc_jurkat-1/jrc_jurkat-1.n5/groundtruth_0003/",
    ),
    RawSources(
        "jrc_macrophage-2",
        (
            f"{cosem_raw_dir}/Macrophage_FS80_Cell2_4x4x4nm/Cryo_FS80_Cell2 4x4x4nm.mrc",
            DisplaySettings(ContrastLimits(0.843, 0.917)),
        ),
        (
            f"{hess_raw_dir}/6. Macrophage_FS80_Cell2_4x4x4nm/Macrophage_FS80_Cell2 4x4x4nm 16bit.mrc",
            DisplaySettings(ContrastLimits(0.297, 0.436), invertColormap=True),
        ),
        pred=dir_glob(
            "/groups/cosem/cosem/ackermand/forDavis/jrc_macrophage-2/jrc_macrophage-2.n5"
        ),
        groundTruth="/nrs/cosem/bennettd/groundtruth/jrc_macrophage-2/jrc_macrophage-2.n5/groundtruth_0003/",
    ),
    RawSources(
        "jrc_sum159-1",
        (
            f"{hess_raw_dir}/8. SUM159_WT45_Cell2_4x4x4nm/SUM159_WT45_Cell2_Cryo_20171009_4x4x4nm.mrc",
            DisplaySettings(ContrastLimits(0.706, 0.864)),
        ),
        (
            f"{hess_raw_dir}/8. SUM159_WT45_Cell2_4x4x4nm/SUM159_WT45_Cell2_Cryo_20171009_4x4x4nm 16bit.mrc",
            DisplaySettings(ContrastLimits(0.241, 0.334), invertColormap=True),            
        ),
        groundTruth="/nrs/cosem/bennettd/groundtruth/jrc_sum159-1/jrc_sum159-1.n5/groundtruth_0003/"
    ),
    RawSources(
        "jrc_ctl-id8-1",
        None,
        (
            f"{hess_raw_dir}/9. TCell_on_cancer_4x4x4nm/TCell_on_cancer_Atlas1_4x4x4nm 16bit.mrc",
            DisplaySettings(ContrastLimits(0.166, 0.265), invertColormap=True),
        ),
    ),
    RawSources(
        "jrc_fly-fsb-1",
        None,
        (
            f"{hess_raw_dir}/4. Fly Fan Shaped Body[Column1-9]_Z0519-11_4x4x4nm/FB-Z0519-11 4x4x4nm 16bit.mrc",
            DisplaySettings(ContrastLimits(0.03433, 0.0509), invertColormap=True),
        ),
    ),
    RawSources(
        "jrc_fly-acc-calyx-1",
        None,
        (
            "/groups/cosem/cosem/bennettd/imports/jrc_fly-acc-calyx-1/jrc_fly-acc-calyx-1.n5/aligned_uint16",
            DisplaySettings(ContrastLimits(0.02499, 0.04994), invertColormap=True),
        ),
    ),
    RawSources(
        "jrc_hela-1",
        (
            f"{cosem_raw_dir}/HeLa_Cell1_8x8x8nm/Aubrey_17-7_17_Cell1_D05-10_8x8x8nm.mrc",
            DisplaySettings(ContrastLimits(.39, .56), invertColormap=True),
        ),
        pred=dir_glob(
            "/groups/cosem/cosem/ackermand/forDavis/jrc_hela-1/jrc_hela-1.n5"
        ),
    ),
    RawSources(
        "jrc_choroid-plexus-2",
        (
            "/nrs/cosem/bennettd/Choroid-Plexus_8x8x8nm/Choroid-Plexus_8x8x8nm.n5/volumes/raw/s0/",
            DisplaySettings(ContrastLimits(0, 1)),
        ),
        pred=dir_glob(
            "/groups/cosem/cosem/ackermand/forDavis/jrc_choroid-plexus-2/jrc_choroid-plexus-2.n5"
        ),
    ),
    RawSources(
        "jrc_cos7-11",
        None,
        (
            f"/nrs/cosem/bennettd/COS7_Cell11_8x8x8nm/SIFTalignTrans-invert.n5/volumes/raw/",
            DisplaySettings(ContrastLimits(0, 1)),
        ),
        lm=('/nrs/saalfeld/john/projects/cosem/COS7_Cell11/LoadID277_Cell11_light.n5/PALM_488', '/nrs/saalfeld/john/projects/cosem/COS7_Cell11/LoadID277_Cell11_light.n5/PALM_532')
    ),
)

def makeVolumeSource(
    datasetName: str,
    volumeName: str,
    path: str,
    displaySettings: DisplaySettings,
    transform: SpatialTransform,
    contentType: str,
    description: str,
    version: str = "0",
    tags: Optional[Tuple[str]] = None,
) -> Optional[VolumeSource]:

    try: 
        arr: xarray.DataArray = DataArrayFromFile(path)
    except zarr.errors.PathNotFoundError as e:
        print(f'Could not access an array at {path}')
        return None

    dimensions: Tuple[int, ...] = arr.shape
    containerType = infer_container_type(path)
    return VolumeSource(
        path=path,
        datasetName=datasetName,
        name=volumeName,
        dataType=str(arr.dtype),
        dimensions=dimensions,
        displaySettings=displaySettings,
        transform=transform,
        contentType=contentType,
        containerType=containerType,
        description=description,
        version=version,
        tags=tags)


def process_raw_sources(
    raw_sources: Sequence[RawSources], class_info: Mapping[str, Tuple[str, str]]
) -> List[Union[VolumeSource, MeshSource]]:
    results: List[Union[VolumeSource, MeshSource, None]] = []
    for raw_source in raw_sources:
        datasetName = raw_source.datasetName
        base_transform = base_transforms[datasetName]

        uint8 = raw_source.uint8
        if uint8 is not None:
            volumeName = "fibsem-uint8"
            path, displaySettings = uint8
            description, _ = class_info[volumeName]
            results.append(
                makeVolumeSource(
                    datasetName=datasetName,
                    volumeName=volumeName,
                    path=path,
                    displaySettings=displaySettings,
                    transform=base_transform,
                    contentType="em",
                    description=description,
                )
            )

        uint16 = raw_source.uint16
        if uint16 is not None:
            volumeName = "fibsem-uint16"
            description, _ = class_info[volumeName]
            path, displaySettings = uint16
            results.append(
                makeVolumeSource(
                    datasetName=datasetName,
                    volumeName=volumeName,
                    path=path,
                    displaySettings=displaySettings,
                    transform=base_transform,
                    contentType="em",
                    description=description,
                )
            )

        pred = raw_source.pred
        if pred is not None:
            volume_paths = pred
            for vp in volume_paths:
                volumeName = Path(vp).name
                className, contentType = get_classname_and_content_type(volumeName)
                description, color = class_info[className]
                displaySettings = DisplaySettings(
                    contrastLimits=ContrastLimits(0, 1), color=color
                )
                
                # the 8nm datasets yield predictions that are 4nm grid spacing
                if base_transform.scale[0] == 8.0:
                    transform = scale(.5, base_transform)
                else:
                    transform = base_transform
                 
                results.append(
                    makeVolumeSource(
                        datasetName,
                        volumeName=volumeName,
                        path=vp,
                        displaySettings=displaySettings,
                        transform=transform,
                        contentType=contentType,
                        description=description,
                    )
                )
        gt = raw_source.groundTruth
        if gt is not None:
            path = gt
            volumeName = "gt"
            description, color = class_info[volumeName]
            displaySettings = DisplaySettings(ContrastLimits(0, 1), color=color)
            results.append(
                makeVolumeSource(
                    datasetName,
                    volumeName=volumeName,
                    path=gt,
                    displaySettings=displaySettings,
                    transform=scale(ground_truth_scaling, base_transform),
                    contentType="segmentation",
                    description=description,
                )
            )
        meshes = raw_source.meshes
        for mp in meshes:
            meshName = Path(mp).name
            results.append(MeshSource(path=mp, 
                                      name=meshName, 
                                      datasetName=datasetName,
                                      format='neuroglancer_precomputed_mesh'))
    
    return list(filter(lambda v: v is not None, results))


def upsert_sources_to_db(sources: Sequence[Union[VolumeSource, MeshSource]]):
    un = "root"
    pw = "root"
    db_name = "cosem"
    # use the datasetName/volumeName as the id for the document
    volumeSources = []
    meshSources = []
    for s in sources:
        f = asdict(s)
        if isinstance(s, VolumeSource):
            f.update({"_id": f["datasetName"] + "/" + f["name"]})
            volumeSources.append(f)
        elif isinstance(s, MeshSource):
            f.update({"_id": f["datasetName"] + "/" + f["name"]})
            meshSources.append(f)
        else:
            f"Object of type {type(f)} cannot be inserted into the database"

    # insert each element in the list into the `datasets` collection on our MongoDB instance
    with MongoClient(f"mongodb://{un}:{pw}@{db_name}.int.janelia.org") as client:        
        # this clearly sucks
        db = client["sources"]["VolumeSource"]
        operations = [
            ReplaceOne(filter={"_id": doc["_id"]}, replacement=doc, upsert=True)
            for doc in volumeSources
        ]
        db.bulk_write(operations)
        
        db = client["sources"]["MeshSource"]
        operations = [
            ReplaceOne(filter={"_id": doc["_id"]}, replacement=doc, upsert=True)
            for doc in meshSources
        ]
        db.bulk_write(operations)
        
        return True


def upsert_views_to_db(dataset_views: Sequence[DatasetView]):
    un = "root"
    pw = "root"
    db_name = "cosem"
    flat = list(map(asdict, dataset_views))
    # use the datasetName/volumeName as the id for the document
    [f.update({"_id": f["datasetName"] + "/" + f["name"]}) for f in flat]

    # insert each element in the list into the `datasets` collection on our MongoDB instance
    with MongoClient(f"mongodb://{un}:{pw}@{db_name}.int.janelia.org") as client:
        db = client["sources"]["DatasetView"]
        operations = [
            ReplaceOne(filter={"_id": doc["_id"]}, replacement=doc, upsert=True)
            for doc in flat
        ]
        result = db.bulk_write(operations)
        return result

@click.command()
@click.option("-v", "--views", required=False, type=bool)
@click.option("-s", "--sources", required=False, type=bool)
def db_update_cli(views: bool, sources: bool):
    result = None
    if sources:        
        volume_sources = process_raw_sources(raw_sources, class_info)
        print(f'Updating database with {len(volume_sources)} sources')
        result = upsert_sources_to_db(volume_sources)
    
    if views:
        dataset_views = get_views_from_google_docs(credfile, view_sheet_name, view_sheet_page)
        print(f'Updating view database with {len(dataset_views)} views')
        result = upsert_views_to_db((dataset_views))        
    
    return result
if __name__ == "__main__":
    db_update_cli()    



import s3fs
from dask import bag
import click
from pathlib import Path
from dask.diagnostics import ProgressBar
from typing import Sequence, Optional, Dict
from dataclasses import dataclass
from ..io import fwalk
from functools import partial

STAGES = ("dev", "prod", "val")


def iterput(
    sources: Sequence[str],
    dests: Sequence[str],
    tags: Optional[Sequence[Optional[Dict]]] = None,
    profile: Optional[str] = None,
    overwrite: bool = True,
):
    """
    Given a sequence of sources, dests, and tags, save each source to dest with a tag.    
    """
    fs = s3fs.S3FileSystem(profile=profile)
    if tags == None:
        tags = (None,) * len(sources)
    for source, dest, tag in zip(sources, dests, tags):
        wrote = False
        try:
            if overwrite:
                fs.put(source, dest)
                wrote = True
            else:
                if not fs.exists(dest):
                    fs.put(source, dest)
                    wrote = True
            if tag is not None and wrote:
                fs.put_tags(dest, tag)
        except OSError as err:
            print(f'Something went wrong copying {source} to {dest}: {err}')
    return True


def s3put(
    dest_root: str,
    source_path: str,
    endswith: Optional[str] = "",
    tags: Optional[Dict] = None,
    partition_size: int=None,
):
    sources = tuple(map(str, fwalk(source_path, endswith)))
    dests = tuple(
        str(Path(dest_root) / Path(f).relative_to(source_path)) for f in sources
    )

    source_bag = bag.from_sequence(sources, partition_size=partition_size)
    dest_bag = bag.from_sequence(dests, partition_size=partition_size)
    tag_bag = bag.from_sequence((tags,) * len(sources), partition_size=partition_size)

    return bag.map_partitions(iterput, source_bag, dest_bag, tag_bag)


@click.command()
@click.argument("source_paths", required=True, nargs=-1)
@click.option("-b", "--bucket", required=True, type=str)
@click.option("-ew", "--endswith", required=True, type=str)
@click.option("-vt", "--version-tag", required=True, type=str)
@click.option("-st", "--stage-tag", required=True, type=str)
@click.option("-dvt", "--developer-tag", required=True, type=str)
@click.option("-pt", "--project-tag", required=True, type=str)
@click.option("-dt", "--description-tag", required=True, type=str)
def s3put_cli(
    source_paths,
    bucket,
    endswith,
    version_tag,
    stage_tag,
    developer_tag,
    project_tag,
    description_tag,
):
    for source_path in source_paths:
        dest_root = Path(bucket) / Path(source_path).stem
        assert stage_tag in STAGES
        tags = {
            "VERSION": version_tag,
            "DEVELOPER": developer_tag,
            "STAGE": stage_tag,
            "PROJECT": project_tag,
            "DESCRIPTION": description_tag,
        }
        s3put(dest_root, source_path, endswith=endswith, tags=tags).compute()
    return 0


@dataclass
class TransferPlan:
    destination: str
    source_root: str
    profile: str
    partition_size: int = 1000
    
    def prepare_source(self, source_file: str):
        return f'{self.destination}/{Path(source_file).relative_to(self.source_root)}'
    
    def prepare_transfer(self, sources: Sequence[str],  overwrite: bool):
        source_bag = bag.from_sequence(sources, partition_size=self.partition_size)
        dest_bag = source_bag.map(self.prepare_source)
        result = bag.map_partitions(partial(iterput, profile=self.profile, overwrite=overwrite), source_bag, dest_bag)
        return result
   


if __name__ == "__main__":
    pbar = ProgressBar()
    pbar.register()
    s3put_cli()



import argparse
from glob import glob
from pathlib import Path
from fst.io.ingest import pad_arrays, arrays_from_delayed
from fst.io import read, access, rmtree_parallel, create_arrays
import numpy as np
import os
import logging
import numcodecs
from fst.distributed import bsub_available, get_jobqueue_cluster, get_cluster
from toolz import partition_all
from typing import Tuple, List, Union

import time
import dask
import dask.array as da
from fst.pyramid import lazy_pyramid, get_downsampled_offset
import zarr

OUTPUT_DTYPES = ("same", "uint8", "uint16")
OUTPUT_FMTS = ("n5",)

program_name = "dat_to_n5.py"

grid_spacing_unit = "nm"
max_chunksize = 1024
compressor = numcodecs.GZip(level=-1)
# all channels will the stored as subgroups under root_group_path
root_group_path = Path("volumes/raw/")
# raw data are stored z | c y x, we will split images into two channels along the channel axis
channel_dim = 0
downscale_factor = 2

# set up logging
logger = logging.getLogger(program_name)
c_handler = logging.StreamHandler()
c_formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
c_handler.setFormatter(c_formatter)
logger.addHandler(c_handler)
logger.setLevel(logging.INFO)


def prepare_data(path: Union[str, list]) -> Tuple[list, list]:
    """
    Takes a path to a collection of .dat files, specified either as single path or a list of individual files,
    and generates list of dask arrays representing the dataset.
    """
    if isinstance(path, str):
        fnames = sorted(glob(path + "*.dat"))
    elif isinstance(path, list):
        fnames = sorted(path)
    else:
        raise ValueError(f"Path variable should be string or list, not {type(path)}")

    logger.info(f"Preparing {len(fnames)} images...")
    data_eager = read(fnames, lazy=False)
    meta, shapes, dtypes = [], [], []
    # build lists of metadata for each image
    for d in data_eager:
        meta.append(d.header.__dict__)
        shapes.append(d.shape)
        dtypes.append(d.dtype)

    data = arrays_from_delayed(read(fnames, lazy=True), shapes=shapes, dtypes=dtypes)
    return data, meta


def read_grid_spacing(metadata: list) -> dict:
    """
    Report the grid spacing of the data in nanometers, given the metadata for an entire
    FIBSEM dataset. The lateral resolution can be inferred from the `PixelSize` attribute of a single
    image, while the axial resolution is an estimate based on the total change in working distance of the SEM
    beam divided by the total number of slices.
    """
    lateral = metadata[0]["PixelSize"]
    axial = abs((metadata[0]["WD"] - metadata[-1]["WD"]) / len(metadata)) / 1e-6
    return {"z": axial, "y": lateral, "x": lateral}


def read_minmax(array_paths: Union[List[str], List[Path]]) -> np.array:
    minmax = np.zeros((len(array_paths), 2))
    for ind, pth in enumerate(array_paths):
        lims = np.array([np.NaN, np.NaN])
        try:
            lims = read(pth).attrs.get("rawHistogramBounds", np.array([np.NaN, np.NaN]))
        except ValueError:
            pass
        minmax[ind] = lims
    return minmax


def calc_minmax(data: list) -> list:
    @dask.delayed
    def minmax(v):
        return np.array((v.min((1, 2)), v.max((1, 2))))

    return list(map(minmax, data))


def prepare_pyramids(
    data, fill_values, chunks, reduction, downscale_factor, output_dtype
) -> Tuple[list, list]:
    padded, padding = pad_arrays(data, constant_values=fill_values)
    stacked = da.stack(padded).rechunk(chunks)
    if output_dtype != "same":
        stacked = (stacked - fill_values.reshape(1, -1, 1, 1)).astype(output_dtype)
    scaling_factors = (1, 1, downscale_factor, downscale_factor)
    pyramid = lazy_pyramid(stacked, reduction, scaling_factors, preserve_dtype=True)
    return pyramid, padding


def change_dtype(data: list, output_dtype: str, offset: np.array) -> list:
    """
    Lazy histogram-preserving datatype adjustment of a collection of array-likes.
    Signed datatypes (int8, int16) are converted to their unsigned counterparts (uint8, uint16) by upcasting to signed type with
    higher precision, shifting all values by a constant, then downcasting to the final unsigned datatype. The resulting arrays
    have a global minimum of 0, with the original min-max distance.
    """

    @dask.delayed
    def adjuster(arr, upcast, offset, dtype):
        assert arr.ndim == offset.ndim
        return (arr.astype(upcast) - offset).astype(dtype)

    if output_dtype == "same":
        return data
    elif output_dtype == "uint8":
        assert (data[0].dtype == "int8") or (data[0].dtype == ">i1")
        upcast = "int16"
    elif output_dtype == "uint16":
        assert (data[0].dtype == "int16") or (data[0].dtype == ">i2")
        upcast = "int32"

    return [
        da.from_delayed(
            adjuster(
                d, upcast=upcast, offset=offset.reshape(-1, 1, 1), dtype=output_dtype
            ),
            dtype=output_dtype,
            shape=d.shape,
        )
        for d in data
    ]


@dask.delayed
def save_pyramid_slicewise(arr, z_index, scale_factors, reduction, depth, output_path):
    pyramid = lazy_pyramid(arr, reduction, scale_factors)[:depth]
    for ind_l, pyr in enumerate(pyramid):
        layer = pyr.data.compute(scheduler="synchronous")
        for ind_ch, ch in enumerate(layer):
            dataset_path = Path(output_path) / f"ch{ind_ch}" / f"s{ind_l}"
            sink = access(dataset_path, mode="a")
            sink[z_index] = ch
    return 0


def dat_to_n5(
    source: str,
    dest: str,
    num_workers: Union[str, int],
    multiscale_levels: int,
    datatype: str,
    output_chunks: tuple,
):
    start_time = time.time()
    num_workers = int(num_workers)
    levels = np.arange(0, multiscale_levels + 1)
    output_fmt = Path(dest).suffix[1:]

    if output_fmt not in OUTPUT_FMTS:
        raise NotImplementedError(
            f"Cannot write a chunked store using format {output_fmt}. Try one of {OUTPUT_FMTS}"
        )

    if datatype not in OUTPUT_DTYPES:
        raise ValueError(f"Datatype must be one of {OUTPUT_DTYPES}")

    client = get_cluster(threads_per_worker=1)
    logger.info(f"Observe progress at {client.cluster.dashboard_link}")

    data, metadata = prepare_data(source)
    grid_spacing = read_grid_spacing(metadata)
    grid_spacing_n5 = (grid_spacing["x"], grid_spacing["y"], grid_spacing["z"])
    num_channels = data[0].shape[channel_dim]

    root_group = access(Path(dest) / root_group_path, mode="a")
    root_group.attrs.put({"metadata": metadata})
    dataset_paths = [
        Path(dest) / root_group_path / f"ch{ch}" for ch in range(num_channels)
    ]

    minmax = read_minmax(dataset_paths)

    if np.any(np.isnan(minmax)):
        client.cluster.scale(num_workers)
        logger.info("Calculating minimum and maximum values of the input data...")
        minmax_per_image = np.array(client.compute(calc_minmax(data), sync=True))
        client.cluster.scale(0)
        minmax = np.array((minmax_per_image.min((0, 1)), minmax_per_image.max((0, 1))))
    else:
        logger.info(
            "Using previously calculated minimum and maximum values of the input data..."
        )

    data_adjusted = change_dtype(data, output_dtype=datatype, offset=minmax[0])
    padded, padding = pad_arrays(data_adjusted, constant_values=0)
    logger.info(f"Assembled dataset with shape {(len(padded), *padded[0].shape)}")

    sample_pyramid = lazy_pyramid(
        np.take(padded[0], 0, channel_dim),
        np.mean,
        (downscale_factor, downscale_factor),
    )[: levels[-1]]

    scales = []
    for l in sample_pyramid:
        scale = list(l.scale_factors)
        scales.append([*scale, 1])

    logger.info("Initializing storage for multiple pyramid levels....")
    for ind_d, dp in enumerate(dataset_paths):
        chunks = [output_chunks] * len(padded)
        shapes, names, dtypes, array_attrs = [], [], [], []
        group_attrs = {
            "downsamplingFactors": scales,
            "padding": padding,
            "createdBy": program_name,
            "rawHistogramBounds": minmax[ind_d].tolist(),
            "datatypeChange": datatype != "same",
        }
        for ind_l, level in enumerate(sample_pyramid):
            shapes.append([len(padded), *level.shape])
            names.append(f"s{ind_l}")
            dtypes.append(level.dtype)
            array_attrs.append(
                {
                    "downsamplingFactors": scales[ind_l],
                    "pixelResolution": {
                        "dimensions": np.multiply(
                            grid_spacing_n5, scales[ind_l]
                        ).tolist(),
                        "units": grid_spacing_unit,
                    },
                    "offset": get_downsampled_offset(scales[ind_l]).tolist(),
                }
            )
        logger.info(f"Initializing {dp}...")
        create_arrays(
            path=dp,
            names=names,
            shapes=shapes,
            dtypes=dtypes,
            compressors=[compressor for n in names],
            chunks=chunks,
            group_attrs=group_attrs,
            array_attrs=array_attrs,
        )

    logger.info("Begin saving data...")

    to_store = [
        save_pyramid_slicewise(
            arr=sl,
            z_index=z_idx,
            scale_factors=(1, downscale_factor, downscale_factor),
            reduction=np.mean,
            depth=levels[-1],
            output_path=f"{dest}/{root_group_path}",
        )
        for (z_idx, sl) in enumerate(padded)
    ]

    partition_size = 10000
    if len(to_store) < partition_size:
        partition_size = len(to_store)

    partitions = partition_all(partition_size, to_store)

    for ind_p, p in enumerate(partitions):
        client.cluster.scale(num_workers)
        logger.info(
            f"Begin saving planes {partition_size * ind_p}:{partition_size * ind_p + len(p)}"
        )
        client.compute(p, sync=True)
        client.cluster.scale(0)

    elapsed_time = time.time() - start_time
    logger.info(f"Save completed in {elapsed_time} s")
    return 0


def main():
    parser = argparse.ArgumentParser(
        description="Save a sequence of images to a chunked store."
    )
    parser.add_argument(
        "-s",
        "--source",
        help="Files to process. Must be either a single directory (e.g., `/data/` "
        "or a wild-card expansion of a single directory (e.g., `/data/*.dat`). "
        "Files will be sorted by filename.",
        required=True,
        nargs="+",
    )
    parser.add_argument(
        "-d",
        "--dest",
        help="The chunked store to create with the input files. Supported chunked store"
        f" formats: {OUTPUT_FMTS} formats are supported",
        required=True,
    )

    parser.add_argument(
        "-nw",
        "--num_workers",
        help="The number of workers to use for distributed computation",
        default=8,
    )

    parser.add_argument(
        "-ml",
        "--multiscale_levels",
        help="The number of multiscale levels to create, in addition to full resolution. E.g., if ml=6 (the default), "
        "downscaled data will be saved with downscaling factors of 2, 4, 8, 16, 32 in addition to full-resolution.",
        default=6,
    )

    parser.add_argument(
        "-dt",
        "--datatype",
        help="The datatype to use for the output. Supported datatypes are `same`, "
        '`uint8`, and `uint16`. Default value is "same", which preserves the '
        "original datatype",
        default="same",
    )

    args = parser.parse_args()
    return dat_to_n5(
        source=args.source,
        dest=args.dest,
        num_workers=args.num_workers,
        multiscale_levels=args.multiscale_levels,
        datatype=args.datatype,
        output_chunks=(1, max_chunksize, max_chunksize),
    )


if __name__ == "__main__":
    main()



from dataclasses import dataclass, asdict
from typing import Any, Dict, Optional, List, Sequence
import tensorstore as ts
import numpy as np
import json
from pathlib import Path
from xarray.core.dataarray import DataArray 

DRIVERS = {'n5', 'neuroglancer_precomputed'}
KVSTORE_DRIVERS = {'file', 'gcs'}

@dataclass
class StripNullFields():
    def asdict(self):
        result = {}
        for k, v in self.__dict__.items():
            if v is not None:
                if hasattr(v, 'asdict'):
                    result[k] = v.asdict()
                elif isinstance(v, list):
                    result[k] = []
                    for element in v:
                        if hasattr(element, 'asdict'):
                            result[k].append(element.asdict)
                        else:
                            result[k].append(element)
                else:
                    result[k] = v
        return result

@dataclass
class KVStore(StripNullFields):
    driver: str
    path: str


@dataclass
class ImageScaleMetadata(StripNullFields):
    size: Sequence[int]
    resolution: Sequence[int]
    encoding: str
    chunk_size: Optional[Sequence[int]] = None
    chunk_sizes: Optional[Sequence[int]] = None
    key: Optional[str] = None
    voxel_offset: Optional[Sequence[int]] = None
    sharding: Optional[Dict[str, Any]] = None


@dataclass
class MultiscaleMetadata(StripNullFields):
    data_type: str
    num_channels: int
    type: str


@dataclass
class TensorStoreSpec(StripNullFields):
    driver: str
    kvstore: KVStore
    path: str
    scale_index: Optional[int] = 0
    scale_metadata: Optional[ImageScaleMetadata] = None
    multiscale_metadata: Optional[MultiscaleMetadata] = None


@dataclass
class PrecomputedMetadata:
    at_type: str
    volume_type: str
    data_type: str
    num_channels: str
    scales: Sequence[ImageScaleMetadata]


def parse_info(text: str) -> PrecomputedMetadata:
    json_data = json.loads(text)
    at_type = json_data['@type']
    volume_type = json_data['type']
    data_type = json_data['data_type']
    num_channels = json_data['num_channels']
    scales = [ImageScaleMetadata(**scale) for scale in json_data['scales']]

    return PrecomputedMetadata(at_type = at_type, 
                               volume_type=volume_type, 
                               data_type=data_type,
                               num_channels=num_channels,
                               scales=scales)

def PrecomputedFromDataArray(dataarray, 
                             path, 
                             encoding, 
                             volume_type, 
                             num_channels=1,
                             voxel_offset=None,
                             key=None, 
                             chunk_size=None,
                             **kwargs):
    assert len(dataarray.coords) == dataarray.ndim
    if voxel_offset is None:
       voxel_offset = (0,) * dataarray.ndim
        
    resolution = [abs(float(a[1].values - a[0].values)) for a in dataarray.coords.values()]
    tsa = TensorStoreArray(driver='neuroglancer_precomputed',
                           path=Path(path).parts[-1],
                           kvstore_driver='file',
                           kvstore_path=str(Path(path).parent),
                           encoding = encoding,
                           volume_type=volume_type,
                           num_channels=num_channels,
                           voxel_offset=voxel_offset,
                           resolution=resolution,
                           key=key,
                           template=dataarray,
                           chunk_size=chunk_size)
    print(tsa.spec['scale_metadata'])
    return tsa.open(**kwargs)

#todo: formally distinguish arrays from groups/containers
class TensorStoreArray:
    def __init__(
        self,
        driver,
        path,
        kvstore_driver,
        kvstore_path,
        encoding,
        num_channels,
        volume_type,
        voxel_offset,
        resolution,
        key=None,
        dtype=None,
        size=None,
        template=None,
        chunk_size=None,
        sharding=None,
        scale_index=None,
    ):
        '''
        A (relatively) friendly interface to tensorstore arrays. 
        '''

        if template is not None:
            if size is not None:
                raise ValueError("Must supply either size or template but not both.")
            if dtype is not None:
                raise ValueError("Must supply either dtype or template but not both.")
            
            size = template.shape
            dtype = template.dtype.name
            if chunk_size is None:
                if hasattr(template, "chunks"):
                    chunk_size = [c[0] for c in template.chunks]                

        if chunk_size is None:
            raise ValueError("Required argument `chunk_size` not supplied.")

        if dtype != 'uint8' and encoding == 'jpeg':
            raise ValueError(f'JPEG encoding only works for uint8 arrays. Your array is {dtype}')
        
        scale_metadata = ImageScaleMetadata(
            chunk_size=chunk_size,
            size=size,
            voxel_offset=voxel_offset,
            resolution=resolution,
            key=key,
            encoding=encoding,
            sharding=sharding,
        )

        kvstore = KVStore(driver=kvstore_driver, path=kvstore_path)

        multiscale_metadata = MultiscaleMetadata(
            data_type=dtype, num_channels=num_channels, type=volume_type
        )

        self.spec = TensorStoreSpec(driver, kvstore, path, scale_index=scale_index,scale_metadata=scale_metadata, multiscale_metadata=multiscale_metadata).asdict()
        
        
    def open(self, **kwargs):
        return ts.open(self.spec, **kwargs)


@dataclass
class NicerTensorStore:
    spec: Dict[str, Any]
    open_kwargs: Dict[str, Any]
    def __getitem__(self, slices):
        return ts.open(spec=self.spec, **self.open_kwargs).result()[slices]
    
    def __setitem__(self, slices, values):
        ts.open(spec=self.spec, **self.open_kwargs).result()[ts.d['channel'][0]][slices] = values
        return None

def prepare_tensorstore_from_pyramid(pyr: Sequence[DataArray], level_names: Sequence[str], jpeg_quality: int, output_chunks: Sequence[int], root_container_path: Path):
    store_arrays = []
    #sharding = {'@type': 'neuroglancer_uint64_sharded_v1',
    #       'preshift_bits': 9,
    #        'hash': 'identity',
    #        'minishard_index_encoding': 'gzip',
    #       'minishard_bits': 6,
    #       'shard_bits': 15}
    
    for p, ln in zip(pyr, level_names):
        res = [abs(float(p.coords[k][1] - p.coords[k][0])) for k in p.dims]
        spec: Dict[str, Any] = {'driver': 'neuroglancer_precomputed',
        'kvstore': {'driver': 'file', 
                    'path': str(Path(root_container_path).parent)},
        'path': root_container_path.parts[-1],
        'scale_metadata': {'size': p.shape,
        'resolution': res,
        'encoding': 'jpeg',
        'jpeg_quality': jpeg_quality,
        #'sharding': sharding,
        'chunk_size': output_chunks,
        'key': ln,
        'voxel_offset': (0, 0, 0)},
        'multiscale_metadata': {'data_type': p.dtype.name,
        'num_channels': 1,
        'type': 'image'}}
        try: 
            ts.open(spec=spec, open=True).result()
        except ValueError:
            try:
                ts.open(spec=spec, create=True).result()
            except ValueError:
                ts.open(spec=spec, create=True, delete_existing=True).result()

        nicer_array = NicerTensorStore(spec=spec, open_kwargs={'write': True})
        store_arrays.append(nicer_array)
    return store_arrays


import zarr
from typing import Union, Any, List
from typing_extensions import TypedDict
from fst.io import read
from pathlib import Path
import json


def jsonify(obj: Union[str, zarr.hierarchy.group, zarr.hierarchy.array]) -> dict:
    """
    Recursively serialize the structure + metadata of an n5/zarr group or array to json
    """

    class Node(TypedDict):
        attrs: str
        nodes: List

    val: Node = {"attrs": "", "nodes": []}

    if isinstance(obj, str):
        try:
            obj = read(obj)
        except ValueError:
            return dict(val)

    result = {}

    attrs_path = Path(obj.store.path) / Path(obj.path) / "attributes.json"
    attrs = json.loads(attrs_path.read_text())
    val["attrs"] = attrs
    key = obj.basename
    result[key] = val
    if hasattr(obj, "values"):
        result[key]["nodes"] = [jsonify(child) for child in obj.values()]

    return result


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Represent a zarr/n5 hierarchy as JSON"
    )
    parser.add_argument("path")
    args = parser.parse_args()
    print(jsonify(args.path))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fibsem_tools',
 'fibsem_tools.attrs',
 'fibsem_tools.io',
 'fibsem_tools.scripts']

package_data = \
{'': ['*']}

install_requires = \
['distributed>=2.30.1,<3.0.0',
 'h5py>=3.1.0,<4.0.0',
 'pymongo>=3.11.2,<4.0.0',
 'sheetscrape>=0.1.0,<0.2.0',
 'tensorstore>=0.1.8,<0.2.0']

setup_kwargs = {
    'name': 'fibsem-tools',
    'version': '0.1.0',
    'description': 'Tools for processing FIBSEM datasets',
    'long_description': None,
    'author': 'Davis Vann Bennett',
    'author_email': 'davis.v.bennett@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from dataclasses import asdict, dataclass
from fst.io.tensorstore import prepare_tensorstore_from_pyramid
from fst.attrs.attrs import CompositeArrayAttrs, DatasetView, MeshSource, makeMultiscaleGroupAttrs
from pathlib import Path
from typing import Any, Callable, Dict, Tuple, Optional, List, TypeVar
import numpy as np
import dask 
dask.config.set({'optimization.fuse.ave-width': 50})
from xarray import DataArray
from distributed import Client
from fst.attrs import (
    DatasetIndex,
    VolumeSource,
    VolumeIngest,
    MultiscaleSpec,
    VolumeStorageSpec,
    CompositeArrayAttrs,
)
from fst.pyramid import lazy_pyramid, mode_reduce, blocked_pyramid, blocked_store
from dask_janelia import auto_cluster
from fst.io import access_multiscale, read
from tqdm import tqdm
import click
from pymongo import MongoClient
from dataclasses import replace
from fsspec import filesystem

def infer_container_type(path: str) -> str:
    if Path(path).suffix == ".mrc":
        containerType = "mrc"
    elif any(map(lambda v: (Path(v).suffix == ".n5"), Path(path).parts)):
        containerType = "n5"
    elif any(map(lambda v: (Path(v).suffix == ".precomputed"), Path(path).parts)):
        containerType = "precomputed"
    else:
        raise ValueError(f"Could not infer container type from path {path}")

    return containerType


def save_dataset_metadata(source: VolumeSource, index: DatasetIndex):
    index.volumes[source.path] = source
    index.to_json(index_file)
    return 0


def flip_axis(arr: DataArray, axis: str) -> DataArray:
    arr2 = arr.copy()
    idx = arr2.dims.index(axis)
    arr2.data = np.flip(arr2.data, idx)
    return arr2


reducers: Dict[str, Callable[[Any], Any]] = {"mean": np.mean, "mode": mode_reduce}
mutations: Dict[str, Callable[[Any], Any]] = {"flip_y": lambda v: flip_axis(v, "y")}
ndim = 3
access_modes = ('write', 'metadata')
save_chunks = (96 * 2,) * ndim
output_chunks = (96,) * ndim
scale_factors = (2,) * ndim
blocked_pyramid_size = (96 * 4, -1, -1)
num_workers = 100
output_dir = "/nrs/cosem/davis/s3"

# mongodb variables
un = "root"
pw = "root"
addr = "cosem.int.janelia.org"
db = "sources"
mongo_addr = f"mongodb://{un}:{pw}@{addr}"


@dataclass
class MultiscaleSavePlan:
    overwrite: bool
    save_chunks: Tuple[int, ...]
    output_chunks: Tuple[int, ...]
    blocked_pyramid_size: Tuple[int, ...]
    jpegQuality: int = 90

T = TypeVar('T')

def typed_list_from_mongodb(mongo_addr: str, db: str, cls: T, query: Dict[str, str]) -> List[T]:
    with MongoClient(mongo_addr) as client:
        retrieved = client[db][cls.__name__].find(query)
    results = []
    for r in retrieved:
        r.pop('_id')
        results.append(cls(**r))
    return results


def makeVolumeIngest(source: VolumeSource, output_path: str) -> VolumeIngest:
    mutation: Optional[str] = None
    reduction = "mean"
    dataPath = ""
    containerPath = ""

    if source.dataType == "uint8" and source.contentType == "em":
        containerType = "precomputed"
        dataPath = ""
        containerPath = str(
            Path(output_path)
            / source.datasetName
            / "neuroglancer" 
            / "em" 
            / f"{source.name}.precomputed"
        )
    else:
        containerType = "n5"
        containerPath = str(
            Path(output_path) / source.datasetName / f"{source.datasetName}.n5"
        )
        if source.contentType == "em":
            dataPath = str(Path("em") / source.name)
        elif source.contentType in {"prediction", "gt", "segmentation", "analysis"}:
            dataPath = str(Path("labels") / source.name)
        else:
            raise ValueError(
                f"content type {source.contentType} cannot be dispatched to a path"
            )
    storageSpec = VolumeStorageSpec("file", containerType, containerPath, dataPath)

    if source.contentType != "em" and source.datasetName in ("jrc_hela-1", "jrc_hela-2", 'jrc_hela-3', 'jrc_jurkat-1', 'jrc_macrophage-2'):
        mutation = "flip_y"

    if source.contentType == "segmentation":
        reduction = "mode"
    multiscaleSpec = MultiscaleSpec(reduction, 5, (2, 2, 2))

    ingest = VolumeIngest(source, multiscaleSpec, storageSpec, mutation)
    return ingest


def build_index(URL: str):
    try:
        fs_protocol, root = URL.split('://')
    except ValueError:
        print(f'Your input {URL} could not be split into a protocol and a path')
        raise
    fs = filesystem(fs_protocol)
    dataset_name = Path(root).name

    n5_paths = tuple(filter(fs.isdir, fs.glob(root +  '/*n5*/*/*')))
    precomputed_paths = tuple(filter(fs.isdir, fs.glob(root +  '/neuroglancer/em/*')))
    mesh_paths = tuple(filter(fs.isdir, fs.glob(root +  '/neuroglancer/mesh/*')))

    output_mesh_sources = [MeshSource(str(Path(mp).relative_to(root)), Path(mp).stem, dataset_name, 'neuroglancer_legacy_mesh') for mp in mesh_paths]

    volume_paths = (*n5_paths, *precomputed_paths)
    volume_path_stems = tuple(Path(s).stem for s in volume_paths)
    
    query = {"datasetName": dataset_name}
    db_volume_sources: List[VolumeSource] = typed_list_from_mongodb(mongo_addr, db, VolumeSource, query)
    if len(db_volume_sources) < 1:
        raise ValueError(f"No sources found in the database using query {query}")
    
    db_source_dict = {s.name: s for s in db_volume_sources}
    output_volume_sources: List[VolumeSource] = []
    for sk, sv in db_source_dict.items():
        if sk not in volume_path_stems:
            print(f'Warning: could not find an extant volume on the filesystem matching this VolumeSource from the database: {sv}. This volume will not be added to the dataset index.')
        else:
            input_source: VolumeSource = replace(sv)
            vol_path = volume_paths[volume_path_stems.index(sk)]
            input_source.path = str(Path(vol_path).relative_to(root))
            input_source.containerType = infer_container_type(vol_path)
            output_volume_sources.append(input_source)

    db_views: List[DatasetView] = []
    db_views = typed_list_from_mongodb(mongo_addr, db, DatasetView, query)
    
    accepted_views = []
    for v in db_views: 
        missing = set(v.volumeKeys) - set(db_source_dict.keys())
        if len(missing) > 0:
            print(f'This view contains volumes: {missing} that could not be found in the volume source database and thus will not be included in the index: {v}')
        else:
            accepted_views.append(v)

    index = DatasetIndex(name=dataset_name, volumes=output_volume_sources, meshes=output_mesh_sources, views=accepted_views)

    return index


def ingest_source(ingest: VolumeIngest, save_plan: MultiscaleSavePlan, num_workers: int):
    source = ingest.source

    # this controls the order of linearized dimensional quantities like pyramid_block_size
    axis_order = slice(None)

    # the default mutation is the identity transformation
    print(f'Using mutation `{ingest.mutation}`')
    mutation = mutations.get(ingest.mutation, lambda v: v)
    darr: DataArray = mutation(source.toDataArray())
    array_attrs = CompositeArrayAttrs.fromDataArray(darr)
    darr.attrs.update(asdict(array_attrs))
    darr.data = darr.data.rechunk(save_plan.save_chunks)
    if ingest.storageSpec.containerType == "precomputed":
        darr = darr.T
        axis_order = slice(-1, None, -1)

    reducer = reducers[ingest.multiscaleSpec.reduction]
    multiscale_depth = ingest.multiscaleSpec.depth
    scale_factors = ingest.multiscaleSpec.factors

    # full_path = Path(output_dir) / Path(source.datasetName) / source.destPath

    pyr = lazy_pyramid(
        darr, reduction=reducer, scale_factors=scale_factors, max_depth=multiscale_depth
    )
    level_names = [f"s{idx}" for idx in range(len(pyr))]
    container_path = Path(output_dir) / ingest.storageSpec.containerPath

    if ingest.storageSpec.containerType == "n5":
        group_path = ingest.storageSpec.dataPath
        print(f"Preparing the store {Path(container_path) / group_path}")
        group_attrs = asdict(makeMultiscaleGroupAttrs(darr.name, pyr, level_names))
        store_group, store_arrays = access_multiscale(
            container_path,
            group_path,
            pyr,
            array_paths=level_names,
            array_chunks=save_plan.output_chunks,
            attr_factory=lambda v: asdict(CompositeArrayAttrs.fromDataArray(v)),
        )
        store_group.attrs.put(group_attrs)
    elif ingest.storageSpec.containerType == "precomputed":
        print(f"Preparing the store {container_path}")
        if darr.dtype.name not in {"uint8"}:
            raise ValueError("Only uint8 supported at this time")

        store_group = None
        store_arrays = prepare_tensorstore_from_pyramid(
            pyr,
            level_names,
            save_plan.jpegQuality,
            save_plan.output_chunks,
            container_path,
        )

    if save_plan.overwrite:
        with auto_cluster(walltime="8:00") as clust, Client(clust) as cl:
            print(cl.cluster.dashboard_link)
            pyr_b = blocked_pyramid(
                darr,
                block_size=save_plan.blocked_pyramid_size[axis_order],
                reduction=reducer,
                scale_factors=scale_factors[axis_order],
                max_depth=multiscale_depth,
            )

            stores = blocked_store(pyr_b, store_arrays, chunks=save_plan.output_chunks)
            cl.cluster.scale(num_workers)
            for store in tqdm(stores):
                cl.compute(store).result()


def check_initialized_chunks(path: str) -> bool:
    r = read(path)
    # i check whether nchunks is less than or equal to initialized chunks because it seems that the zarr python library
    # is a little generous with what it considers a chunk, and thus it's possible to have more chunks initialized than total chunks,
    # e.g. if some .nfsblablabla files are floating around from other processes
    return all(i.nchunks <= i.initialized for n,i in r.items())


@click.command()
@click.option("-q", "--query", required=True, type=str)
@click.option("-w", "--workers", required=False, type=int, default=0)
@click.option("--overwrite/--no-overwrite", default=False)
@click.option("--metadata/--no-metadata", default=False)
def ingest_source_cli(query: str, workers: int, overwrite: bool, metadata: bool):
    if overwrite and metadata:
        print('Warning: you have selected overwrite AND metadata operating modes. The metadata operating mode takes precedence, and thus incomplete containers will not be completed.')
        overwrite = False
    query_result = get_sources_from_mongodb(eval(query))
    if len(query_result) == 0:
        raise ValueError(f'Could not find a single dataset using the query {query}')
    sources = query_result
    for source in sources:
        ingest = makeVolumeIngest(source, output_dir)
        plan = MultiscaleSavePlan(overwrite=overwrite, save_chunks=save_chunks, output_chunks=output_chunks, blocked_pyramid_size=blocked_pyramid_size)
        pth = (Path(ingest.storageSpec.containerPath) / ingest.storageSpec.dataPath)
        incomplete = False
        if not pth.exists():
            incomplete = True
            print(f'Nothing found at {pth}. Multiscale volume will be created.')
            if not metadata:
                plan.overwrite = True
        elif ingest.storageSpec.containerType == 'n5' and not metadata:
            incomplete = not check_initialized_chunks(str(pth))
            if incomplete:
                print(f'Incomplete multiscale volume found at {pth}. Multiscale volume will be created.')
                plan.overwrite = True
                
        print(f'Ingesting store at {pth}...')
        ingest_source(ingest, plan, num_workers=workers)
        
if __name__ == '__main__':
    ingest_source_cli()


"""
Functions for reading FIB-SEM data from Harald Hess' proprietary format
Adapted from https://github.com/janelia-cosem/FIB-SEM-Aligner/blob/master/fibsem.py
Copyright (c) 2017, David Hoffman, Davis Bennett
"""

import os
import numpy as np
from typing import Iterable, Union, Sequence
from pathlib import Path
import dask.array as da
from xarray import DataArray

Pathlike = Union[str, Path]

# This value is used to ensure that the endianness of the data is correct
MAGIC_NUMBER = 3555587570
# This is the size of the header, in bytes. Everything beyond this number of bytes is data.
OFFSET = 1024


class FIBSEMHeader(object):
    """Structure to hold header info"""

    def __init__(self, **kwargs):
        self.__dict__ = kwargs

    def update(self, **kwargs):
        """update internal dictionary"""
        self.__dict__.update(kwargs)

    def to_native_types(self):
        for k, v in self.__dict__.items():
            if isinstance(v, np.integer):
                self.__dict__[k] = int(v)
            elif isinstance(v, np.bytes_):
                self.__dict__[k] = v.tostring().decode("utf-8")
            elif isinstance(v, np.ndarray):
                self.__dict__[k] = v.tolist()
            elif isinstance(v, np.floating):
                self.__dict__[k] = float(v)
            else:
                self.__dict__[k] = v


class FIBSEMData(np.ndarray):
    """Subclass of ndarray to attach header data to fibsem data"""

    def __new__(cls, input_array, header=None):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance
        obj.header = header
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return
        self.header = getattr(obj, "header", None)


class _DTypeDict(object):
    """Handle dtype dict manipulations"""

    def __init__(self, names=None, formats=None, offsets=None):
        # initialize internals to empty lists
        self.names = []
        self.formats = []
        self.offsets = []
        if names is not None:
            self.update(names, formats, offsets)

    def update(self, names, formats, offsets):
        """"""
        if isinstance(names, list):
            if len(names) == len(formats) == len(offsets):
                self.names.extend(names)
                self.formats.extend(formats)
                self.offsets.extend(offsets)
            else:
                raise RuntimeError("Lengths are not equal")
        else:
            self.names.append(names)
            self.formats.append(formats)
            self.offsets.append(offsets)

    @property
    def dict(self):
        """Return the dict representation"""
        return dict(names=self.names, formats=self.formats, offsets=self.offsets)

    @property
    def dtype(self):
        """return the dtype"""
        return np.dtype(self.dict)


def _read_header(path):
    # make emtpy header to fill
    header_dtype = _DTypeDict()
    header_dtype.update(
        [
            "FileMagicNum",  # Read in magic number, should be 3555587570
            "FileVersion",  # Read in file version number
            "FileType",  # Read in file type, 1 is Zeiss Neon detectors
            "SWdate",  # Read in SW date
            "TimeStep",  # Read in AI sampling time (including oversampling) in seconds
            "ChanNum",  # Read in number of channels
            "EightBit",  # Read in 8-bit data switch
        ],
        [">u4", ">u2", ">u2", ">S10", ">f8", ">u1", ">u1"],
        [0, 4, 6, 8, 24, 32, 33],
    )
    # read initial header
    with open(path, mode="rb") as fobj:
        base_header = np.fromfile(fobj, dtype=header_dtype.dtype, count=1)
        fibsem_header = FIBSEMHeader(
            **dict(zip(base_header.dtype.names, base_header[0]))
        )
        # now fobj is at position 34, return to 0
        fobj.seek(0, os.SEEK_SET)
        if fibsem_header.FileMagicNum != MAGIC_NUMBER:
            raise RuntimeError(
                f"FileMagicNum should be {MAGIC_NUMBER} but is {fibsem_header.FileMagicNum}"
            )

        _scaling_offset = 36
        if fibsem_header.FileVersion == 1:
            header_dtype.update(
                "Scaling", (">f8", (fibsem_header.ChanNum, 4)), _scaling_offset
            )
        elif fibsem_header.FileVersion in {2, 3, 4, 5, 6}:
            header_dtype.update(
                "Scaling", (">f4", (fibsem_header.ChanNum, 4)), _scaling_offset
            )
        else:
            # Read in AI channel scaling factors, (col#: AI#), (row#: offset, gain, 2nd order, 3rd order)
            header_dtype.update("Scaling", (">f4", (2, 4)), _scaling_offset)

        header_dtype.update(
            ["XResolution", "YResolution"],  # X Resolution  # Y Resolution
            [">u4", ">u4"],
            [100, 104],
        )

        if fibsem_header.FileVersion in {1, 2, 3}:
            header_dtype.update(
                ["Oversampling", "AIDelay"],  # AI oversampling  # Read AI delay (
                [">u1", ">i2"],
                [108, 109],
            )
        else:
            header_dtype.update("Oversampling", ">u2", 108)  # AI oversampling

        header_dtype.update("ZeissScanSpeed", ">u1", 111)  # Scan speed (Zeiss #)
        if fibsem_header.FileVersion in {1, 2, 3}:
            header_dtype.update(
                [
                    "ScanRate",  # Actual AO (scanning) rate
                    "FramelineRampdownRatio",  # Frameline rampdown ratio
                    "Xmin",  # X coil minimum voltage
                    "Xmax",  # X coil maximum voltage
                ],
                [">f8", ">f8", ">f8", ">f8"],
                [112, 120, 128, 136],
            )
            # fibsem_header.Detmin = -10 # Detector minimum voltage
            # fibsem_header.Detmax = 10 # Detector maximum voltage
        else:
            header_dtype.update(
                [
                    "ScanRate",  # Actual AO (scanning) rate
                    "FramelineRampdownRatio",  # Frameline rampdown ratio
                    "Xmin",  # X coil minimum voltage
                    "Xmax",  # X coil maximum voltage
                    "Detmin",  # Detector minimum voltage
                    "Detmax",  # Detector maximum voltage
                    "DecimatingFactor",  # Decimating factor
                ],
                [">f4", ">f4", ">f4", ">f4", ">f4", ">f4", ">u2"],
                [112, 116, 120, 124, 128, 132, 136],
            )

        header_dtype.update(
            [
                "AI1",  # AI Ch1
                "AI2",  # AI Ch2
                "AI3",  # AI Ch3
                "AI4",  # AI Ch4
                "Notes",  # Read in notes
            ],
            [">u1", ">u1", ">u1", ">u1", ">S200"],
            [151, 152, 153, 154, 180],
        )

        if fibsem_header.FileVersion in {1, 2}:
            header_dtype.update(
                [
                    "DetA",  # Name of detector A
                    "DetB",  # Name of detector B
                    "DetC",  # Name of detector C
                    "DetD",  # Name of detector D
                    "Mag",  # Magnification
                    "PixelSize",  # Pixel size in nm
                    "WD",  # Working distance in mm
                    "EHT",  # EHT in kV
                    "SEMApr",  # SEM aperture number
                    "HighCurrent",  # high current mode (1=on, 0=off)
                    "SEMCurr",  # SEM probe current in A
                    "SEMRot",  # SEM scan roation in degree
                    "ChamVac",  # Chamber vacuum
                    "GunVac",  # E-gun vacuum
                    "SEMStiX",  # SEM stigmation X
                    "SEMStiY",  # SEM stigmation Y
                    "SEMAlnX",  # SEM aperture alignment X
                    "SEMAlnY",  # SEM aperture alignment Y
                    "StageX",  # Stage position X in mm
                    "StageY",  # Stage position Y in mm
                    "StageZ",  # Stage position Z in mm
                    "StageT",  # Stage position T in degree
                    "StageR",  # Stage position R in degree
                    "StageM",  # Stage position M in mm
                    "BrightnessA",  # Detector A brightness (
                    "ContrastA",  # Detector A contrast (
                    "BrightnessB",  # Detector B brightness (
                    "ContrastB",  # Detector B contrast (
                    "Mode",
                    # FIB mode: 0=SEM, 1=FIB, 2=Milling, 3=SEM+FIB, 4=Mill+SEM, 5=SEM Drift Correction, 6=FIB Drift Correction, 7=No Beam, 8=External, 9=External+SEM
                    "FIBFocus",  # FIB focus in kV
                    "FIBProb",  # FIB probe number
                    "FIBCurr",  # FIB emission current
                    "FIBRot",  # FIB scan rotation
                    "FIBAlnX",  # FIB aperture alignment X
                    "FIBAlnY",  # FIB aperture alignment Y
                    "FIBStiX",  # FIB stigmation X
                    "FIBStiY",  # FIB stigmation Y
                    "FIBShiftX",  # FIB beam shift X in micron
                    "FIBShiftY",  # FIB beam shift Y in micron
                ],
                [
                    ">S10",
                    ">S18",
                    ">S20",
                    ">S20",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">u1",
                    ">u1",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">u1",
                    ">f8",
                    ">u1",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                ],
                [
                    380,
                    390,
                    700,
                    720,
                    408,
                    416,
                    424,
                    432,
                    440,
                    441,
                    448,
                    456,
                    464,
                    472,
                    480,
                    488,
                    496,
                    504,
                    512,
                    520,
                    528,
                    536,
                    544,
                    552,
                    560,
                    568,
                    576,
                    584,
                    600,
                    608,
                    616,
                    624,
                    632,
                    640,
                    648,
                    656,
                    664,
                    672,
                    680,
                ],
            )
        else:
            header_dtype.update(
                [
                    "DetA",  # Name of detector A
                    "DetB",  # Name of detector B
                    "DetC",  # Name of detector C
                    "DetD",  # Name of detector D
                    "Mag",  # Magnification
                    "PixelSize",  # Pixel size in nm
                    "WD",  # Working distance in mm
                    "EHT",  # EHT in kV
                    "SEMApr",  # SEM aperture number
                    "HighCurrent",  # high current mode (1=on, 0=off)
                    "SEMCurr",  # SEM probe current in A
                    "SEMRot",  # SEM scan roation in degree
                    "ChamVac",  # Chamber vacuum
                    "GunVac",  # E-gun vacuum
                    "SEMShiftX",  # SEM beam shift X
                    "SEMShiftY",  # SEM beam shift Y
                    "SEMStiX",  # SEM stigmation X
                    "SEMStiY",  # SEM stigmation Y
                    "SEMAlnX",  # SEM aperture alignment X
                    "SEMAlnY",  # SEM aperture alignment Y
                    "StageX",  # Stage position X in mm
                    "StageY",  # Stage position Y in mm
                    "StageZ",  # Stage position Z in mm
                    "StageT",  # Stage position T in degree
                    "StageR",  # Stage position R in degree
                    "StageM",  # Stage position M in mm
                    "BrightnessA",  # Detector A brightness (#)
                    "ContrastA",  # Detector A contrast (#)
                    "BrightnessB",  # Detector B brightness (#)
                    "ContrastB",  # Detector B contrast (#)
                    "Mode",
                    # FIB mode: 0=SEM, 1=FIB, 2=Milling, 3=SEM+FIB, 4=Mill+SEM, 5=SEM Drift Correction, 6=FIB Drift Correction, 7=No Beam, 8=External, 9=External+SEM
                    "FIBFocus",  # FIB focus in kV
                    "FIBProb",  # FIB probe number
                    "FIBCurr",  # FIB emission current
                    "FIBRot",  # FIB scan rotation
                    "FIBAlnX",  # FIB aperture alignment X
                    "FIBAlnY",  # FIB aperture alignment Y
                    "FIBStiX",  # FIB stigmation X
                    "FIBStiY",  # FIB stigmation Y
                    "FIBShiftX",  # FIB beam shift X in micron
                    "FIBShiftY",  # FIB beam shift Y in micron
                ],
                [
                    ">S10",
                    ">S18",
                    ">S20",
                    ">S20",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">u1",
                    ">u1",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">u1",
                    ">f4",
                    ">u1",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                ],
                [
                    380,
                    390,
                    410,
                    430,
                    460,
                    464,
                    468,
                    472,
                    480,
                    481,
                    490,
                    494,
                    498,
                    502,
                    510,
                    514,
                    518,
                    522,
                    526,
                    530,
                    534,
                    538,
                    542,
                    546,
                    550,
                    554,
                    560,
                    564,
                    568,
                    572,
                    600,
                    604,
                    608,
                    620,
                    624,
                    628,
                    632,
                    636,
                    640,
                    644,
                    648,
                ],
            )

        if fibsem_header.FileVersion in {5, 6, 7, 8}:
            header_dtype.update(
                [
                    "MillingXResolution",  # FIB milling X resolution
                    "MillingYResolution",  # FIB milling Y resolution
                    "MillingXSize",  # FIB milling X size (um)
                    "MillingYSize",  # FIB milling Y size (um)
                    "MillingULAng",  # FIB milling upper left inner angle (deg)
                    "MillingURAng",  # FIB milling upper right inner angle (deg)
                    "MillingLineTime",  # FIB line milling time (s)
                    "FIBFOV",  # FIB FOV (um)
                    "MillingLinesPerImage",  # FIB milling lines per image
                    "MillingPIDOn",  # FIB milling PID on
                    "MillingPIDMeasured",  # FIB milling PID measured (0:specimen, 1:beamdump)
                    "MillingPIDTarget",  # FIB milling PID target
                    "MillingPIDTargetSlope",  # FIB milling PID target slope
                    "MillingPIDP",  # FIB milling PID P
                    "MillingPIDI",  # FIB milling PID I
                    "MillingPIDD",  # FIB milling PID D
                    "MachineID",  # Machine ID
                    "SEMSpecimenI",  # SEM specimen current (nA)
                ],
                [
                    ">u4",
                    ">u4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">u2",
                    ">u1",
                    ">u1",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">S30",
                    ">f4",
                ],
                [
                    652,
                    656,
                    660,
                    664,
                    668,
                    672,
                    676,
                    680,
                    684,
                    686,
                    689,
                    690,
                    694,
                    698,
                    702,
                    706,
                    800,
                    980,
                ],
            )

        if fibsem_header.FileVersion in {6, 7}:
            header_dtype.update(
                [
                    "Temperature",  # Temperature (F)
                    "FaradayCupI",  # Faraday cup current (nA)
                    "FIBSpecimenI",  # FIB specimen current (nA)
                    "BeamDump1I",  # Beam dump 1 current (nA)
                    "SEMSpecimenI",  # SEM specimen current (nA)
                    "MillingYVoltage",  # Milling Y voltage (V)
                    "FocusIndex",  # Focus index
                    "FIBSliceNum",  # FIB slice #
                ],
                [">f4", ">f4", ">f4", ">f4", ">f4", ">f4", ">f4", ">u4"],
                [850, 854, 858, 862, 866, 870, 874, 878],
            )
        if fibsem_header.FileVersion == 8:
            header_dtype.update(
                [
                    "BeamDump2I",  # Beam dump 2 current (nA)
                    "MillingI",  # Milling current (nA)
                ],
                [">f4", ">f4"],
                [882, 886],
            )
        header_dtype.update("FileLength", ">i8", 1000)  # Read in file length in bytes

        # read header
        header = np.fromfile(fobj, dtype=header_dtype.dtype, count=1)
        fibsem_header = FIBSEMHeader(**dict(zip(header.dtype.names, header[0])))
        fibsem_header.to_native_types()

    return fibsem_header


def _read(path: Union[str, Path]) -> FIBSEMData:
    """

    Read a single .dat file.

    Parameters
    ----------
    path : string denoting a path to a .dat file

    Returns : an instance of FIBSEMData
    -------

    """
    # Load raw_data data file 's' or 'ieee-be.l64' Big-ian ordering, 64-bit long data type
    fibsem_header = _read_header(path)
    # read data
    shape = (
        fibsem_header.YResolution,
        fibsem_header.XResolution,
        fibsem_header.ChanNum,
    )

    if fibsem_header.EightBit == 1:
        dtype = ">u1"
    else:
        dtype = ">i2"

    try:
        raw_data = np.memmap(path, dtype=dtype, mode="r", offset=OFFSET, shape=shape,)
    except ValueError:
        # todo: read what data is available
        raw_data = np.zeros(dtype=dtype, shape=shape)

    raw_data = np.rollaxis(raw_data, 2)
    # Once read into the FIBSEMData structure it will be in memory, not memmap.
    result = FIBSEMData(raw_data, fibsem_header)
    del raw_data
    return result


def read_fibsem(path: Union[str, Path, Iterable[str], Iterable[Path]]):
    """

    Parameters
    ----------
    path : string or iterable of strings representing paths to .dat files.

    Returns a single FIBSEMDataset or an iterable of FIBSEMDatasets, depending on whether a single path or multiple
    paths are supplied as arguments.
    -------

    """
    if isinstance(path, (str, Path)):
        return _read(path)
    elif isinstance(path, Iterable):
        return [_read(p) for p in path]
    else:
        raise ValueError(
            "Path must be an instance of string or pathlib.Path, or iterable of strings / pathlib.Paths"
        )


def _convert_data(fibsem):
    """

    Parameters
    ----------
    fibsem

    Returns
    -------

    """
    # Convert raw_data data to electron counts
    if fibsem.header.EightBit == 1:
        scaled = np.empty(fibsem.shape, dtype=np.int16)
        detector_a, detector_b = fibsem
        if fibsem.header.AI1:
            detector_a = fibsem[0]
            scaled[0] = np.int16(
                fibsem[0]
                * fibsem.header.ScanRate
                / fibsem.header.Scaling[0, 0]
                / fibsem.header.Scaling[0, 2]
                / fibsem.header.Scaling[0, 3]
                + fibsem.header.Scaling[0, 1]
            )
            if fibsem.header.AI2:
                detector_b = fibsem[1]
                scaled[1] = np.int16(
                    fibsem[1]
                    * fibsem.header.ScanRate
                    / fibsem.header.Scaling[1, 0]
                    / fibsem.header.Scaling[1, 2]
                    / fibsem.header.Scaling[1, 3]
                    + fibsem.header.Scaling[1, 1]
                )

        elif fibsem.header.AI2:
            detector_b = fibsem[0]
            scaled[0] = np.int16(
                fibsem[0]
                * fibsem.header.ScanRate
                / fibsem.header.Scaling[0, 0]
                / fibsem.header.Scaling[0, 2]
                / fibsem.header.Scaling[0, 3]
                + fibsem.header.Scaling[0, 1]
            )

    else:
        if fibsem.header.FileVersion in {1, 2, 3, 4, 5, 6}:
            # scaled =

            if fibsem.header.AI1:
                detector_a = (
                    fibsem.header.Scaling[0, 0]
                    + fibsem[0] * fibsem.header.Scaling[0, 1]
                )
                if fibsem.header.AI2:
                    detector_b = (
                        fibsem.header.Scaling[1, 0]
                        + fibsem[1] * fibsem.header.Scaling[1, 1]
                    )
                    if fibsem.header.AI3:
                        detector_c = (
                            fibsem.header.Scaling[2, 0]
                            + fibsem[1] * fibsem.header.Scaling[1, 1]
                        )
        else:
            pass
    return detector_a, detector_b, scaled


def _chunked_fibsem_loader(
    filenames, channel_axis, pad_values=None, concat_axis=0, block_info=None
):
    """
    Load fibsem data and pad if needed. Designed to work with da.map_blocks.
    """
    idx = block_info[None]["chunk-location"]
    output_shape = block_info[None]["chunk-shape"]
    filedata = np.expand_dims(
        np.asanyarray(read_fibsem(filenames[idx[0]])), concat_axis
    )
    pad_width = np.subtract(output_shape, filedata.shape)
    if np.any(pad_width):
        if not pad_values:
            raise ValueError("Data must be padded but no pad values were supplied!")
        padded = []
        pw = np.take(pad_width, [x for x in range(len(pad_width)) if x != channel_axis])
        for ind, channel in enumerate(np.swapaxes(filedata, 0, channel_axis)):
            padded.append(
                np.pad(
                    channel,
                    list(zip(pw * 0, pw)),
                    mode="constant",
                    constant_values=pad_values[ind],
                )
            )

        result = np.stack(padded, channel_axis)
    else:
        result = filedata
    return result


def minmax(filenames, block_info):
    idx = block_info[None]["chunk-location"][0]
    filedata = read_fibsem(filenames[idx])
    return np.expand_dims(np.array([[f.min(), f.max()] for f in filedata]), 0)


def aggregate_fibsem_metadata(fnames):
    headers = [_read_header(f) for f in fnames]
    meta, shapes, dtypes = [], [], []
    # build lists of metadata for each image
    for d in headers:
        if d.EightBit == 1:
            dtype = ">u1"
        else:
            dtype = ">i2"
        meta.append(d.__dict__)
        shapes.append((d.ChanNum, d.YResolution, d.XResolution))
        dtypes.append(dtype)
    return meta, shapes, dtypes


class FibsemDataset:
    def __init__(self, filenames: Sequence):
        """
        Create a representation of a collection of .dat files as a single dataset.
        """
        self.filenames = filenames
        self.metadata, self.shapes, self.dtypes = aggregate_fibsem_metadata(
            self.filenames
        )
        self.axes = {"z": 0, "c": 1, "y": 2, "x": 3}
        self.bounding_shape = {
            k: v
            for k, v in zip(
                self.axes, (len(self.filenames), *np.array(self.shapes).max(0))
            )
        }
        self.extrema = self._get_extrema()
        self.needs_padding = len(set(self.shapes)) > 1
        self.grid_spacing, self.coords = self._get_grid_spacing_and_coords()

    def _get_grid_spacing_and_coords(self):
        lateral = self.metadata[0]["PixelSize"]
        axial = (
            abs((self.metadata[0]["WD"] - self.metadata[-1]["WD"]) / len(self.metadata))
            / 1e-6
        )
        grid_spacing = {"z": axial, "c": 1, "y": lateral, "x": lateral}
        coords = {
            k: np.arange(0, grid_spacing[k] * self.bounding_shape[k], grid_spacing[k])
            for k in grid_spacing
        }
        return grid_spacing, coords

    def _get_extrema(self):
        all_extrema = da.map_blocks(
            minmax,
            self.filenames,
            chunks=((1,) * len(self.filenames), self.bounding_shape["c"], 2),
            dtype=self.dtypes[0],
        )
        result_data = da.stack([all_extrema.min(0)[:, 0], all_extrema.max(0)[:, 1]])
        result = DataArray(
            result_data, dims=("stat", "c"), coords={"stat": ["min", "max"]}
        )
        return result

    def to_dask(self, pad_values=None):
        num_channels = self.bounding_shape["c"]
        if self.needs_padding:
            if pad_values is None:
                raise ValueError("Data must be padded but no pad values were supplied!")
            elif len(pad_values) != num_channels:
                raise ValueError(
                    f"Length of pad values {pad_values} does not match the length of the channel axis ({num_channels})"
                )

        chunks = (
            (1,) * self.bounding_shape["z"],
            *[self.bounding_shape[k] for k in ("c", "y", "x")],
        )
        darr = da.map_blocks(
            _chunked_fibsem_loader,
            self.filenames,
            self.axes["c"],
            pad_values,
            chunks=chunks,
            dtype=self.dtypes[0],
        )
        return darr
